# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
The dummy model encodes the model defined by django in backends.djsite
using SQLAlchemy.
This is done to query the database with more performant ORM of SA.
"""

from sqlalchemy.ext.declarative import declarative_base

from sqlalchemy import (
    Column, Table, ForeignKey, UniqueConstraint,create_engine,
    select, func, join, and_, or_, not_, except_, case, exists,
    text
)

from sqlalchemy.types import (
    Integer, String, DateTime, Float, Boolean, Text,
)
from sqlalchemy.orm import (
    relationship,
    backref,
    sessionmaker,
    foreign, mapper, aliased
)

from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm.attributes import InstrumentedAttribute
from sqlalchemy.sql.elements import Cast
from sqlalchemy.dialects.postgresql import UUID, JSONB, INTEGER, array
# TO COMPILE MY OWN FUNCTIONALITIES:
from sqlalchemy.sql.expression import FunctionElement, cast
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.ext.compiler import compiles


# SETTINGS:
from aiida.common.setup import get_profile_config
from aiida.backends import settings

#EXCEPTIONS
from aiida.common.exceptions import DbContentError, MissingPluginError

# MISC
from aiida.backends.sqlalchemy.models.utils import uuid_func
from aiida.utils import timezone
from aiida.common.datastructures import (calc_states, _sorted_datastates,
                                         sort_states)

Base = declarative_base()

class DbLink(Base):
    __tablename__ = "db_dblink"
    id = Column(Integer, primary_key=True)
    input_id = Column(
        Integer,
        ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
    )
    output_id = Column(
        Integer,
        ForeignKey(
            'db_dbnode.id',
            ondelete="CASCADE",
            deferrable=True,
            initially="DEFERRED"
        )
    )
    type=Column(String(255))
    input = relationship("DbNode", primaryjoin="DbLink.input_id == DbNode.id")
    output = relationship("DbNode", primaryjoin="DbLink.output_id == DbNode.id")
    label = Column(String(255), index=True, nullable=False)

class DbCalcState(Base):
    __tablename__ = "db_dbcalcstate"
    id = Column(Integer, primary_key=True)
    dbnode = relationship(
            'DbNode',
            backref=backref('dbstates', passive_deletes=True)
        )
    state = Column(String(255))
    time = Column(DateTime(timezone=True), default=timezone.now)
    dbnode_id = Column(
            Integer,
            ForeignKey(
                'db_dbnode.id', ondelete="CASCADE",
                deferrable=True, initially="DEFERRED"
            )
        )
    __table_args__ = (
        UniqueConstraint('dbnode_id', 'state'),
    )

class DbAttribute(Base):
    __tablename__ = "db_dbattribute"
    id  = Column(Integer, primary_key = True)
    dbnode_id = Column(Integer, ForeignKey('db_dbnode.id'))
    key = Column(String(255))
    datatype = Column(String(10))
    tval = Column(String, default='')
    fval = Column(Float, default=None, nullable=True)
    ival = Column(Integer, default=None, nullable=True)
    bval = Column(Boolean, default=None, nullable=True)
    dval = Column(DateTime, default=None, nullable = True)



class DbExtra(Base):
    __tablename__ = "db_dbextra"
    id  = Column(Integer, primary_key = True)
    dbnode_id = Column(Integer, ForeignKey('db_dbnode.id'))
    key = Column(String(255))
    datatype = Column(String(10))
    tval = Column(String, default='')
    fval = Column(Float, default=None, nullable=True)
    ival = Column(Integer, default=None, nullable=True)
    bval = Column(Boolean, default=None, nullable=True)
    dval = Column(DateTime, default=None, nullable = True)


class DbComputer(Base):
    __tablename__ = "db_dbcomputer"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    name = Column(String(255), unique=True, nullable=False)
    hostname = Column(String(255))

    description = Column(Text, nullable=True)
    enabled = Column(Boolean)

    transport_type = Column(String(255))
    scheduler_type = Column(String(255))

    transport_params = Column(String(255))
    _metadata = Column('metadata', String(255), default="{}")

    def get_aiida_class(self):
        from aiida.backends.djsite.db.models import DbComputer as DjangoSchemaDbComputer
        djcomputer = DjangoSchemaDbComputer(
            id=self.id, uuid=self.uuid, name=self.name,
            hostname=self.hostname, description=self.description,
            enabled=self.enabled, transport_type=self.transport_type,
            scheduler_type=self.scheduler_type, transport_params=self.transport_params,
            metadata=self._metadata
        )
        return djcomputer.get_aiida_class()

class DbUser(Base):
    __tablename__ = "db_dbuser"

    id = Column(Integer, primary_key=True)
    email = Column(String(254), unique=True, index=True)
    password = Column(String(128))  # Clear text password ?
    first_name = Column(String(254), nullable=True)
    last_name = Column(String(254), nullable=True)
    institution = Column(String(254), nullable=True)

    is_staff = Column(Boolean, default=False)
    is_active = Column(Boolean, default=False)

    last_login = Column(DateTime(timezone=True), default=timezone.now)
    date_joined = Column(DateTime(timezone=True), default=timezone.now)

    def get_aiida_class(self):
        from aiida.backends.djsite.db.models import DbUser as DjangoSchemaDbUser
        djuser = DjangoSchemaDbUser(
            id=self.id,email=self.email, password=self.password,
            first_name=self.first_name, last_name=self.last_name,
            institution=self.institution, is_staff=self.is_staff,
            is_active=self.is_active, last_login=self.last_login,
            date_joined=self.date_joined
        )
        return djuser.get_aiida_class()


table_groups_nodes = Table(
    'db_dbgroup_dbnodes',
    Base.metadata,
    Column('id', Integer, primary_key=True),
    Column(
            'dbnode_id', Integer,
            ForeignKey('db_dbnode.id', deferrable=True, initially="DEFERRED")
        ),
    Column(
            'dbgroup_id', Integer,
            ForeignKey('db_dbgroup.id', deferrable=True, initially="DEFERRED")
        )
)

class DbGroup(Base):
    __tablename__ = "db_dbgroup"

    id = Column(Integer, primary_key=True)

    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    name = Column(String(255), index=True)

    type = Column(String(255), default="", index=True)

    time = Column(DateTime(timezone=True), default=timezone.now)
    description = Column(Text, nullable=True)

    user_id = Column(
            Integer,
            ForeignKey(
                'db_dbuser.id', ondelete='CASCADE',
                deferrable=True, initially="DEFERRED")
            )
    user = relationship('DbUser', backref=backref('dbgroups', cascade='merge'))

    dbnodes = relationship('DbNode', secondary=table_groups_nodes,
                           backref="dbgroups", lazy='dynamic')

    __table_args__ = (
        UniqueConstraint('name', 'type'),
    )

    def __str__(self):
        if self.type:
            return '<DbGroup [type: {}] "{}">'.format(self.type, self.name)
        else:
            return '<DbGroup [user-defined] "{}">'.format(self.name)
    def get_aiida_class(self):
        from aiida.orm.implementation.django.group import Group as DjangoAiidaGroup
        from aiida.backends.djsite.db.models import DbGroup as DjangoSchemaDbGroup
        dbgroup = DjangoSchemaDbGroup(
            id = self.id,
            type = self.type,
            uuid = self.uuid,
            name = self.name,
            time = self.time,
            description  = self.description,
            user_id = self.user_id,
        )

        return DjangoAiidaGroup(dbgroup=dbgroup)



class DbNode(Base):
    __tablename__ = "db_dbnode"
    id = Column(Integer, primary_key=True)
    uuid = Column(UUID(as_uuid=True), default=uuid_func)
    type = Column(String(255), index=True)
    process_type = Column(String(255), index=True)
    label = Column(String(255), index=True, nullable=True)
    description = Column(Text(), nullable=True)
    ctime = Column(DateTime(timezone=True), default=timezone.now)
    mtime = Column(DateTime(timezone=True), default=timezone.now)
    dbcomputer_id = Column(
        Integer,
        ForeignKey('db_dbcomputer.id', deferrable=True, initially="DEFERRED"),
        nullable=True
    )
    dbcomputer = relationship(
        'DbComputer',
        backref=backref('dbnodes', passive_deletes=True)
    )
    user_id = Column(
        Integer,
        ForeignKey('db_dbuser.id', deferrable=True, initially="DEFERRED"),
        nullable=False
    )
    user = relationship('DbUser', backref='dbnodes')

    public = Column(Boolean, default=False)

    nodeversion = Column(Integer, default=1)

    attributes = relationship('DbAttribute', uselist=True, backref='dbnode')
    extras = relationship('DbExtra', uselist=True, backref='dbnode')



    outputs = relationship(
        "DbNode",
        secondary       =   "db_dblink",
        primaryjoin     =   "DbNode.id == DbLink.input_id",
        secondaryjoin   =   "DbNode.id == DbLink.output_id",
        backref         =   backref("inputs", passive_deletes=True),
        passive_deletes =   True
    )

    def get_aiida_class(self):
        """
        Return the corresponding instance of
        :func:`~aiida.orm.implementation.django.node.Node`
        or a subclass return by the plugin loader.

        .. todo::
            The behavior is quite pathetic, creating a django DbNode instance
            to instantiate the aiida instance.
            These means that every time you load Aiida instances with
            the QueryBuilder when using Django as a backend, three instances
            are instantiated for every Aiida instance you load!
            Could be fixed by allowing DbNode from the dummy nodel to be passed
            to AiidaNode's __init__.

        :returns: An instance of the plugin class
        """
        # I need to import the DbNode in the Django model,
        # and instantiate an object that has the same attributes as self.
        from aiida.backends.djsite.db.models import DbNode as DjangoSchemaDbNode
        dbnode = DjangoSchemaDbNode(
            id=self.id, type=self.type, process_type=self.process_type, uuid=self.uuid, ctime=self.ctime,
            mtime=self.mtime, label=self.label, description=self.description, dbcomputer_id=self.dbcomputer_id,
            user_id=self.user_id, public=self.public, nodeversion=self.nodeversion
        )
        return dbnode.get_aiida_class()

    @hybrid_property
    def user_email(self):
        """
        Returns: the email of the user
        """
        return self.user.email

    @user_email.expression
    def user_email(cls):
        """
        Returns: the email of the user at a class level (i.e. in the database)
        """
        return select([DbUser.email]).where(DbUser.id == cls.user_id).label(
            'user_email')


    # Computer name
    @hybrid_property
    def computer_name(self):
        """
        Returns: the of the computer
        """
        return self.dbcomputer.name

    @computer_name.expression
    def computer_name(cls):
        """
        Returns: the name of the computer at a class level (i.e. in the 
        database)
        """
        return select([DbComputer.name]).where(DbComputer.id ==
                                                 cls.dbcomputer_id).label(
            'computer_name')

    # State
    @hybrid_property
    def state(self):
        """
        Return the most recent state from DbCalcState
        """
        if not self.id:
            return None
        all_states = DbCalcState.query.filter(DbCalcState.dbnode_id == self.id).all()
        if all_states:
            #return max((st.time, st.state) for st in all_states)[1]
            return sort_states(((dbcalcstate.state, dbcalcstate.state.value)
                                for dbcalcstate in all_states),
                                use_key=True)[0]
        else:
            return None

    @state.expression
    def state(cls):
        """
        Return the expression to get the 'latest' state from DbCalcState,
        to be used in queries, where 'latest' is defined using the state order
        defined in _sorted_datastates.
        """
        # Sort first the latest states
        whens = {
            v: idx for idx, v
            in enumerate(_sorted_datastates[::-1], start=1)}
        custom_sort_order = case(value=DbCalcState.state,
                                 whens=whens,
                                 else_=100) # else: high value to put it at the bottom

        # Add numerical state to string, to allow to sort them
        states_with_num = select([
            DbCalcState.id.label('id'),
            DbCalcState.dbnode_id.label('dbnode_id'),
            DbCalcState.state.label('state_string'),
            custom_sort_order.label('num_state')
        ]).select_from(DbCalcState).alias()

        # Get the most 'recent' state (using the state ordering, and the min function) for
        # each calc
        calc_state_num = select([
            states_with_num.c.dbnode_id.label('dbnode_id'),
            func.min(states_with_num.c.num_state).label('recent_state')
        ]).group_by(states_with_num.c.dbnode_id).alias()

        # Join the most-recent-state table with the DbCalcState table
        all_states_q = select([
            DbCalcState.dbnode_id.label('dbnode_id'),
            DbCalcState.state.label('state_string'),
            calc_state_num.c.recent_state.label('recent_state'),
            custom_sort_order.label('num_state'),
        ]).select_from(#DbCalcState).alias().join(
            join(DbCalcState, calc_state_num, DbCalcState.dbnode_id == calc_state_num.c.dbnode_id)).alias()

        # Get the association between each calc and only its corresponding most-recent-state row
        subq = select([
            all_states_q.c.dbnode_id.label('dbnode_id'),
            all_states_q.c.state_string.label('state')
        ]).select_from(all_states_q).where(all_states_q.c.num_state == all_states_q.c.recent_state).alias()

        # Final filtering for the actual query
        return select([subq.c.state]).\
            where(
                    subq.c.dbnode_id == cls.id,
                ).\
            label('laststate')




profile = get_profile_config(settings.AIIDADB_PROFILE)


def get_aldjemy_session():
    """
    Use aldjemy to make a session

    .. note:
        Use only in this case. In normal production mode
        it is safer make session explictly because it is more robust
    """
    from aldjemy.core import get_engine
    engine = get_engine()
    _Session = sessionmaker(bind=engine)
    return _Session()


session = get_aldjemy_session()





# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida import work
from aiida.backends.testbase import AiidaTestCase
from aiida.common.utils import classproperty
from aiida.orm.data.int import Int
from aiida.orm.calculation.job.simpleplugins.templatereplacer import TemplatereplacerCalculation
from aiida.work.persistence import ObjectLoader
from aiida.work.job_processes import JobProcess
from aiida.work import Process

from . import utils


class AdditionalParameterCalculation(TemplatereplacerCalculation):
    """
    Subclass of TemplatereplacerCalculation that also defines a use method
    with an additional parameter
    """

    @classproperty
    def _use_methods(cls):
        retdict = TemplatereplacerCalculation._use_methods
        retdict.update({
            'pseudo': {
                'valid_types': Int,
                'additional_parameter': "kind",
                'linkname': cls._get_linkname_pseudo,
                'docstring': (''),
            },
        })
        return retdict

    @classmethod
    def _get_linkname_pseudo(cls, kind):
        """
        Create the linkname based on the additional parameter
        """
        if isinstance(kind, (tuple, list)):
            suffix_string = '_'.join(kind)
        elif isinstance(kind, basestring):
            suffix_string = kind
        else:
            raise TypeError('invalid additional parameter type')

        return '{}_{}'.format('pseudo', suffix_string)


class TestJobProcess(AiidaTestCase):

    def setUp(self):
        super(TestJobProcess, self).setUp()
        self.assertIsNone(Process.current())
        self.runner = utils.create_test_runner()

    def tearDown(self):
        super(TestJobProcess, self).tearDown()
        self.assertIsNone(Process.current())
        self.runner.close()
        self.runner = None
        work.set_runner(None)

    def test_class_loader(self):
        cl = ObjectLoader()
        TemplatereplacerProcess = JobProcess.build(TemplatereplacerCalculation)

    def test_job_process_set_label_and_description(self):
        """
        Verify that calculation label and description get set when passed through inputs
        """
        label = 'test_label'
        description = 'test_description'
        inputs = {
            'options': {
                'computer': self.computer,
                'resources': {
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 1
                },
                'max_wallclock_seconds': 10,
            },
            'label': label,
            'description': description
        }
        process = TemplatereplacerCalculation.process()
        job = process(inputs)

        self.assertEquals(job.calc.label, label)
        self.assertEquals(job.calc.description, description)

    def test_job_process_set_none(self):
        """
        Verify that calculation label and description can be not set.
        """
        inputs = {
            'options': {
                'computer': self.computer,
                'resources': {
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 1
                },
                'max_wallclock_seconds': 10,
            }
        }

        process = TemplatereplacerCalculation.process()
        job = process(inputs)


class TestAdditionalParameterJobProcess(AiidaTestCase):

    def setUp(self):
        super(TestAdditionalParameterJobProcess, self).setUp()
        self.assertIsNone(Process.current())
        self.runner = utils.create_test_runner()

    def tearDown(self):
        super(TestAdditionalParameterJobProcess, self).tearDown()
        self.assertIsNone(Process.current())
        self.runner.close()
        self.runner = None
        work.set_runner(None)

    def test_class_loader(self):
        cl = ObjectLoader()
        AdditionalParameterProcess = JobProcess.build(AdditionalParameterCalculation)

    def test_job_process_with_additional_parameter(self):
        """
        Verify that the additional parameter use_method 'pseudo' is supported
        """
        label = 'test_label'
        description = 'test_description'
        inputs = {
            'options': {
                'computer': self.computer,
                'resources': {
                    'num_machines': 1,
                    'num_mpiprocs_per_machine': 1
                },
                'max_wallclock_seconds': 10,
            },
            'pseudo': {
                'a': Int(1),
                'b': Int(2),
            },
            'label': label,
            'description': description
        }
        process = AdditionalParameterCalculation.process()
        job = process(inputs)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import sys

from aiida.cmdline.baseclass import VerdiCommandWithSubcommands
from aiida.backends.utils import load_dbenv, is_dbenv_loaded
from aiida.common.exceptions import ValidationError


def prompt_for_computer_configuration(computer):
    import inspect, readline
    from aiida.orm.computer import Computer
    from aiida.common.exceptions import ValidationError

    for internal_name, name, desc, multiline in (
            Computer._conf_attributes):
        # Check if I should skip this entry
        shouldcall_name = '_shouldcall_{}'.format(internal_name)
        try:
            shouldcallfunc = dict(inspect.getmembers(
                computer))[shouldcall_name]
            shouldcall = shouldcallfunc()
        except KeyError:
            shouldcall = True
        if not shouldcall:
            # Call cleanup code, if present
            cleanup_name = '_cleanup_{}'.format(internal_name)
            try:
                cleanup = dict(inspect.getmembers(
                    computer))[cleanup_name]
                cleanup()
            except KeyError:
                # No cleanup function: this is not a problem, simply
                # no cleanup is needed
                pass

            # Skip this question
            continue

        getter_name = '_get_{}_string'.format(internal_name)
        try:
            getter = dict(inspect.getmembers(
                computer))[getter_name]
        except KeyError:
            print >> sys.stderr, ("Internal error! "
                                  "No {} getter defined in Computer".format(getter_name))
            sys.exit(1)
        previous_value = getter()

        setter_name = '_set_{}_string'.format(internal_name)
        try:
            setter = dict(inspect.getmembers(
                computer))[setter_name]
        except KeyError:
            print >> sys.stderr, ("Internal error! "
                                  "No {} setter defined in Computer".format(setter_name))
            sys.exit(1)

        valid_input = False
        while not valid_input:
            if multiline:
                newlines = []
                print "=> {}: ".format(name)
                print "   # This is a multiline input, press CTRL+D on a"
                print "   # empty line when you finish"

                try:
                    for l in previous_value.splitlines():
                        while True:
                            readline.set_startup_hook(lambda:
                                                      readline.insert_text(l))
                            input_txt = raw_input()
                            if input_txt.strip() == '?':
                                print ["  > {}".format(descl) for descl
                                       in "HELP: {}".format(desc).split('\n')]
                                continue
                            else:
                                newlines.append(input_txt)
                                break

                    # Reset the hook (no default text printed)
                    readline.set_startup_hook()

                    print "   # ------------------------------------------"
                    print "   # End of old input. You can keep adding     "
                    print "   # lines, or press CTRL+D to store this value"
                    print "   # ------------------------------------------"

                    while True:
                        input_txt = raw_input()
                        if input_txt.strip() == '?':
                            print "\n".join(["  > {}".format(descl) for descl
                                             in "HELP: {}".format(desc).split('\n')])
                            continue
                        else:
                            newlines.append(input_txt)
                except EOFError:
                    # Ctrl+D pressed: end of input.
                    pass

                input_txt = "\n".join(newlines)

            else:  # No multiline
                readline.set_startup_hook(lambda: readline.insert_text(
                    previous_value))
                input_txt = raw_input("=> {}: ".format(name))
                if input_txt.strip() == '?':
                    print "HELP:", desc
                    continue

            try:
                setter(input_txt)
                valid_input = True
            except ValidationError as e:
                print >> sys.stderr, "Invalid input: {}".format(e.message)
                print >> sys.stderr, "Enter '?' for help".format(e.message)


class Computer(VerdiCommandWithSubcommands):
    """
    Setup and manage computers to be used

    This command allows to list, add, modify and configure computers.
    """

    def __init__(self):
        """
        A dictionary with valid commands and functions to be called.
        """
        super(Computer, self).__init__()

        self.valid_subcommands = {
            'list': (self.computer_list, self.complete_none),
            'show': (self.computer_show, self.complete_computers),
            'setup': (self.computer_setup, self.complete_none),
            'update': (self.computer_update, self.complete_computers),
            'enable': (self.computer_enable, self.complete_computers),
            'disable': (self.computer_disable, self.complete_computers),
            'rename': (self.computer_rename, self.complete_computers),
            'configure': (self.computer_configure, self.complete_computers),
            'test': (self.computer_test, self.complete_computers),
            'delete': (self.computer_delete, self.complete_computers),
        }

    def complete_computers(self, subargs_idx, subargs):
        if not is_dbenv_loaded():
            load_dbenv()
        computer_names = self.get_computer_names()
        print computer_names
        return "\n".join(computer_names)

    def computer_list(self, *args):
        """
        List available computers
        """
        import argparse

        if not is_dbenv_loaded():
            load_dbenv()
        from aiida.orm.computer import Computer as AiiDAOrmComputer

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='List the computers in the database.')
        # The default states are those that are shown if no option is given
        parser.add_argument('-C', '--color', action='store_true',
                            help="Use colors to help visualizing the different categories",
                            )
        parser.add_argument('-o', '--only-usable', action='store_true',
                            help="Show only computers that are usable (i.e., "
                                 "configured for the given user and enabled)",
                            )
        parser.add_argument('-p', '--parsable', action='store_true',
                            help="Show only the computer names, one per line, "
                                 "without any other information or string.",
                            )
        parser.add_argument('-a', '--all', action='store_true',
                            help="Show also disabled or unconfigured computers",
                            )
        parser.set_defaults(also_disabled=False)
        parsed_args = parser.parse_args(args)
        use_colors = parsed_args.color
        only_usable = parsed_args.only_usable
        parsable = parsed_args.parsable
        all_comps = parsed_args.all

        computer_names = self.get_computer_names()

        if use_colors:
            color_id = 90  # Dark gray
            color_id = None  # Default color
            if color_id is not None:
                start_color = "\x1b[{}m".format(color_id)
                end_color = "\x1b[0m"
            else:
                start_color = ""
                end_color = ""
        else:
            start_color = ""
            end_color = ""

        if not parsable:
            print "{}# List of configured computers:{}".format(
                start_color, end_color)
            print ("{}# (use 'verdi computer show COMPUTERNAME' "
                   "to see the details){}".format(start_color, end_color))
        if computer_names:
            for name in sorted(computer_names):
                computer = AiiDAOrmComputer.get(name)

                # color_id = 90 # Dark gray
                # color_id = 34 # Blue

                is_configured = computer.is_user_configured(self.backend.users.get_automatic_user())
                is_user_enabled = computer.is_user_enabled(self.backend.users.get_automatic_user())

                is_usable = False  # True if both enabled and configured

                if not all_comps:
                    if not is_configured or not is_user_enabled or not computer.is_enabled():
                        continue

                if computer.is_enabled():
                    if is_configured:
                        configured_str = ""
                        if is_user_enabled:
                            symbol = "*"
                            color_id = None
                            enabled_str = ""
                            is_usable = True
                        else:
                            symbol = "x"
                            color_id = 31  # Red
                            enabled_str = "[DISABLED for this user]"
                    else:
                        symbol = "x"
                        color_id = 90  # Dark gray
                        enabled_str = ""
                        configured_str = " [unconfigured]"
                else:  # GLOBALLY DISABLED
                    symbol = "x"
                    color_id = 31  # Red
                    if is_configured and not is_user_enabled:
                        enabled_str = " [DISABLED globally AND for this user]"
                    else:
                        enabled_str = " [DISABLED globally]"
                    if is_configured:
                        configured_str = ""
                    else:
                        configured_str = " [unconfigured]"

                if use_colors:
                    if color_id is not None:
                        start_color = "\x1b[{}m".format(color_id)
                        bold_sequence = "\x1b[1;{}m".format(color_id)
                        nobold_sequence = "\x1b[0;{}m".format(color_id)
                    else:
                        start_color = "\x1b[0m"
                        bold_sequence = "\x1b[1m"
                        nobold_sequence = "\x1b[0m"
                    end_color = "\x1b[0m"
                else:
                    start_color = ""
                    end_color = ""
                    bold_sequence = ""
                    nobold_sequence = ""

                if parsable:
                    print "{}{}{}".format(start_color, name, end_color)
                else:
                    if (not only_usable) or is_usable:
                        print "{}{} {}{}{} {}{}{}".format(
                            start_color, symbol,
                            bold_sequence, name, nobold_sequence,
                            enabled_str, configured_str, end_color)

        else:
            print "# No computers configured yet. Use 'verdi computer setup'"

    def computer_show(self, *args):
        """
        Show information on a given computer
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.common.exceptions import NotExistent

        if len(args) != 1:
            print >> sys.stderr, ("after 'computer show' there should be one "
                                  "argument only, being the computer name.")
            sys.exit(1)
        try:
            computer = self.get_computer(name=args[0])
        except NotExistent:
            print >> sys.stderr, "No computer in the DB with name {}.".format(
                args[0])
            sys.exit(1)
        print computer.full_text_info

    def computer_update(self, *args):
        """
        Update an existing computer
        """
        import argparse
        from aiida.common.exceptions import NotExistent

        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.orm.computer import Computer

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Update a computer')
        # The default states are those that are shown if no option is given
        parser.add_argument('computer_name', help="The name of the computer")
        parsed_args = parser.parse_args(args)
        computer_name = parsed_args.computer_name

        try:
            computer = Computer.get(computer_name)
        except NotExistent:
            print "No computer {} was found".format(computer_name)
            sys.exit(1)

        calculation_on_computer = computer.get_calculations_on_computer()

        if calculation_on_computer:
            # Note: this is an artificial comment.
            # If you comment the following lines, you will be able to overwrite
            # the old computer anyway, at your own risk.
            print "You cannot modify a computer, after you run some calculations on it."
            print "Disable this computer and set up a new one."
            sys.exit(1)

        print "*" * 75
        print "WARNING! Modifying existing computer with name '{}'".format(computer_name)
        print "Are you sure you want to continue? The UUID will remain the same!"
        print "Continue only if you know what you are doing."
        print "If you just want to rename a computer, use the 'verdi computer rename'"
        print "command. In most cases, it is better to create a new computer."
        print "Moreover, if you change the transport, you must also reconfigure"
        print "each computer for each user!"
        print "*" * 75
        print "Press [Enter] to continue, or [Ctrl]+C to exit."
        raw_input()

        prompt_for_computer_configuration(computer)

        try:
            computer.store()
        except ValidationError as e:
            print "Unable to store the computer: {}. Exiting...".format(e.message)
            sys.exit(1)

        print "Computer '{}' successfully updated.".format(computer_name)
        print "pk: {}, uuid: {}".format(computer.pk, computer.uuid)
        print "(Note: machine_dependent transport parameters cannot be set via "
        print "the command-line interface at the moment)"

        print "OK"
        pass

    def computer_setup(self, *args):
        """
        Setup a new or existing computer
        """
        import readline

        if len(args) != 0:
            print >> sys.stderr, ("after 'computer setup' there cannot be any "
                                  "argument.")
            sys.exit(1)

        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.common.exceptions import NotExistent, ValidationError
        from aiida.orm.computer import Computer as AiidaOrmComputer

        print "At any prompt, type ? to get some help."
        print "---------------------------------------"

        # get the new computer name
        readline.set_startup_hook(lambda: readline.insert_text(previous_value))
        input_txt = raw_input("=> Computer name: ")
        if input_txt.strip() == '?':
            print "HELP:", "The computer name"
        computer_name = input_txt.strip()

        try:
            computer = self.get_computer(name=computer_name)
            print "A computer called {} already exists.".format(computer_name)
            print "Use 'verdi computer update' to update it, and be careful if"
            print "you really want to modify a database entry!"
            print "Now exiting..."
            sys.exit(1)
        except NotExistent:
            computer = AiidaOrmComputer(name=computer_name)
            print "Creating new computer with name '{}'".format(computer_name)

        prompt_for_computer_configuration(computer)

        try:
            computer.store()
        except ValidationError as e:
            print "Unable to store the computer: {}. Exiting...".format(e.message)
            sys.exit(1)

        print "Computer '{}' successfully stored in DB.".format(computer_name)
        print "pk: {}, uuid: {}".format(computer.pk, computer.uuid)
        print "Note: before using it with AiiDA, configure it using the command"
        print "  verdi computer configure {}".format(computer_name)
        print "(Note: machine_dependent transport parameters cannot be set via "
        print "the command-line interface at the moment)"

    def computer_rename(self, *args):
        """
        Rename a computer
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.common.exceptions import (
            NotExistent, UniquenessError, ValidationError)

        try:
            oldname = args[0]
            newname = args[1]
        except IndexError:
            print >> sys.stderr, "Pass as parameters the old and the new name"
            sys.exit(1)

        if oldname == newname:
            print >> sys.stderr, "The initial and final name are the same."
            sys.exit(1)

        try:
            computer = self.get_computer(name=oldname)
        except NotExistent:
            print >> sys.stderr, "No computer exists with name '{}'".format(
                oldname)
            sys.exit(1)

        try:
            computer.set_name(newname)
            computer.store()
        except ValidationError as e:
            print >> sys.stderr, "Invalid input! {}".format(e.message)
            sys.exit(1)
        except UniquenessError as e:
            print >> sys.stderr, ("Uniqueness error encountered! Probably a "
                                  "computer with name '{}' already exists"
                                  "".format(newname))
            print >> sys.stderr, "(Message was: {})".format(e.message)
            sys.exit(1)

        print "Computer '{}' renamed to '{}'".format(oldname, newname)

    def computer_configure(self, *args):
        """
        Configure the authentication information for a given computer
        """
        if not is_dbenv_loaded():
            load_dbenv()

        import readline
        import inspect

        from aiida.common.exceptions import (NotExistent, ValidationError, MultipleObjectsError)
        from aiida.common.utils import get_configured_user_email

        import argparse

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Configure a computer for a given AiiDA user.')
        # The default states are those that are shown if no option is given
        parser.add_argument('-u', '--user', type=str, metavar='EMAIL',
                            help="Configure the computer for the given AiiDA user (otherwise, configure the current default user)",
                            )
        parser.add_argument('computer', type=str,
                            help="The name of the computer that you want to configure")

        parsed_args = parser.parse_args(args)

        user_email = parsed_args.user
        computername = parsed_args.computer

        try:
            computer = self.get_computer(name=computername)
        except NotExistent:
            print >> sys.stderr, "No computer exists with name '{}'".format(
                computername)
            sys.exit(1)
        if user_email is None:
            user = self.backend.users.get_automatic_user()
        else:
            try:
                user = self.backend.users.get(email=user_email)
            except (NotExistent, MultipleObjectsError) as e:
                print >> sys.stderr, ("{}".format(e))
                sys.exit(1)

        try:
            authinfo = self.backend.authinfos.get(computer=computer, user=user)
        except NotExistent:
            authinfo = self.backend.authinfos.create(computer=computer, user=user)
        old_authparams = authinfo.get_auth_params()

        Transport = computer.get_transport_class()

        print ("Configuring computer '{}' for the AiiDA user '{}'".format(
            computername, user.email))

        print "Computer {} has transport of type {}".format(computername,
                                                            computer.get_transport_type())

        if user.email != get_configured_user_email():
            print "*" * 72
            print "** {:66s} **".format("WARNING!")
            print "** {:66s} **".format(
                "  You are configuring a different user.")
            print "** {:66s} **".format(
                "  Note that the default suggestions are taken from your")
            print "** {:66s} **".format(
                "  local configuration files, so they may be incorrect.")
            print "*" * 72

        valid_keys = Transport.get_valid_auth_params()

        default_authparams = {}
        for k in valid_keys:
            if k in old_authparams:
                default_authparams[k] = old_authparams.pop(k)
        if old_authparams:
            print ("WARNING: the following keys were previously in the "
                   "authorization parameters,")
            print "but have not been recognized and have been deleted:"
            print ", ".join(old_authparams.keys())

        print ""
        print "Note: to leave a field unconfigured, leave it empty and press [Enter]"

        # I strip out the old auth_params that are not among the valid keys
        new_authparams = {}

        for k in valid_keys:
            key_set = False
            while not key_set:
                try:
                    converter_name = '_convert_{}_fromstring'.format(k)
                    try:
                        converter = dict(inspect.getmembers(
                            Transport))[converter_name]
                    except KeyError:
                        print >> sys.stderr, ("Internal error! "
                                              "No {} defined in Transport {}".format(
                            converter_name, computer.get_transport_type()))
                        sys.exit(1)

                    if k in default_authparams:
                        readline.set_startup_hook(lambda:
                                                  readline.insert_text(str(default_authparams[k])))
                    else:
                        # Use suggestion only if parameters were not already set
                        suggester_name = '_get_{}_suggestion_string'.format(k)
                        try:
                            suggester = dict(inspect.getmembers(
                                Transport))[suggester_name]
                            suggestion = suggester(computer)
                            readline.set_startup_hook(lambda:
                                                      readline.insert_text(suggestion))
                        except KeyError:
                            readline.set_startup_hook()

                    txtval = raw_input("=> {} = ".format(k))
                    if txtval:
                        new_authparams[k] = converter(txtval)
                    key_set = True
                except ValidationError as e:
                    print "Error in the inserted value: {}".format(e.message)

        if not valid_keys:
            print "There are no special keys to be configured. Configuration completed."

        authinfo.set_auth_params(new_authparams)
        authinfo.store()
        print "Configuration stored for your user on computer '{}'.".format(
            computername)

    def computer_delete(self, *args):
        """
        Configure the authentication information for a given computer

        Does not delete the computer if there are calculations that are using
        it.
        """
        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.common.exceptions import (
            NotExistent, InvalidOperation)
        from aiida.orm.computer import delete_computer

        if len(args) != 1:
            print >> sys.stderr, ("after 'computer delete' there should be one "
                                  "argument only, being the computer name.")
            sys.exit(1)

        computername = args[0]

        try:
            computer = self.get_computer(name=computername)
        except NotExistent:
            print >> sys.stderr, "No computer exists with name '{}'".format(
                computername)
            sys.exit(1)

        try:
            delete_computer(computer)
        except InvalidOperation as e:
            print >> sys.stderr, e.message
            sys.exit(1)

        print "Computer '{}' deleted.".format(computername)

    def computer_test(self, *args):
        """
        Test the connection to a computer.

        It tries to connect, to get the list of calculations on the queue and
        to perform other tests.
        """
        import argparse
        import traceback

        if not is_dbenv_loaded():
            load_dbenv()

        from aiida.common.exceptions import NotExistent

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Test a remote computer')
        # The default states are those that are shown if no option is given
        parser.add_argument('-u', '--user', type=str, metavar='EMAIL',
                            dest='user_email',
                            help="Test the connection for a given AiiDA user."
                                 "If not specified, uses the current "
                                 "default user.",
                            )
        parser.add_argument('-t', '--traceback', action='store_true',
                            help="Print the full traceback in case an exception "
                                 "is raised",
                            )
        parser.add_argument('computer', type=str,
                            help="The name of the computer that you "
                                 "want to test")

        parsed_args = parser.parse_args(args)

        user_email = parsed_args.user_email
        computername = parsed_args.computer
        print_traceback = parsed_args.traceback

        try:
            computer = self.get_computer(name=computername)
        except NotExistent:
            print >> sys.stderr, "No computer exists with name '{}'".format(
                computername)
            sys.exit(1)

        if user_email is None:
            user = self.backend.users.get_automatic_user()
        else:
            user_list = self.backend.users.find(email=user_email)
            # If no user is found
            if not user_list:
                print >> sys.stderr, ("No user with email '{}' in the "
                                      "database.".format(user_email))
                sys.exit(1)
            user = user_list[0]

        print "Testing computer '{}' for user {}...".format(computername,
                                                            user.email)
        try:
            authinfo = computer.get_authinfo(user)
        except NotExistent:
            print >> sys.stderr, ("User with email '{}' is not yet configured "
                                  "for computer '{}' yet.".format(
                user.email, computername))
            sys.exit(1)

        warning_string = None
        if not authinfo.enabled:
            warning_string = ("** NOTE! Computer is disabled for the "
                              "specified user!\n   Do you really want to test it? [y/N] ")
        if not computer.is_enabled():
            warning_string = ("** NOTE! Computer is disabled!\n"
                              "   Do you really want to test it? [y/N] ")
        if warning_string:
            answer = raw_input(warning_string)
            if not (answer == 'y' or answer == 'Y'):
                sys.exit(0)

        s = authinfo.computer.get_scheduler()
        t = authinfo.get_transport()

        ## STARTING TESTS HERE
        num_failures = 0
        num_tests = 0

        try:
            print "> Testing connection..."
            with t:
                s.set_transport(t)
                num_tests += 1
                for test in [self._computer_test_get_jobs,
                             self._computer_create_temp_file]:
                    num_tests += 1
                    try:
                        succeeded = test(transport=t, scheduler=s,
                                         authinfo=authinfo)
                    except Exception as e:
                        print "* The test raised an exception!"
                        if print_traceback:
                            print "** Full traceback:"
                            # Indent
                            print "\n".join(["   {}".format(l) for l
                                             in traceback.format_exc().splitlines()])
                        else:
                            print "** {}: {}".format(e.__class__.__name__,
                                                     e.message)
                            print ("** (use the --traceback option to see the "
                                   "full traceback)")
                        succeeded = False

                    if not succeeded:
                        num_failures += 1

            if num_failures:
                print "Some tests failed! ({} out of {} failed)".format(
                    num_failures, num_tests)
            else:
                print "Test completed (all {} tests succeeded)".format(
                    num_tests)
        except Exception as e:
            print "** Error while trying to connect to the computer! Cannot "
            print "   perform following tests, stopping."
            if print_traceback:
                print "** Full traceback:"
                # Indent
                print "\n".join(["   {}".format(l) for l
                                 in traceback.format_exc().splitlines()])
            else:
                print "{}: {}".format(e.__class__.__name__, e.message)
                print ("(use the --traceback option to see the "
                       "full traceback)")
            succeeded = False

    def _computer_test_get_jobs(self, transport, scheduler, authinfo):
        """
        Internal test to check if it is possible to check the queue state.

        :note: exceptions could be raised

        :param transport: an open transport
        :param scheduler: the corresponding scheduler class
        :param authinfo: the authinfo object (from which one can get
          computer and aiidauser)
        :return: True if the test succeeds, False if it fails.
        """
        print "> Getting job list..."
        found_jobs = scheduler.getJobs(as_dict=True)
        # For debug
        # for jid, data in found_jobs.iteritems():
        #    print jid, data['submission_time'], data['dispatch_time'], data['job_state']
        print "  `-> OK, {} jobs found in the queue.".format(len(found_jobs))
        return True

    def _computer_create_temp_file(self, transport, scheduler, authinfo):
        """
        Internal test to check if it is possible to create a temporary file
        and then delete it in the work directory

        :note: exceptions could be raised

        :param transport: an open transport
        :param scheduler: the corresponding scheduler class
        :param authinfo: the AuthInfo object (from which one can get
          computer and aiidauser)
        :return: True if the test succeeds, False if it fails.
        """
        import tempfile
        import datetime
        import os

        file_content = "Test from 'verdi computer test' on {}".format(
            datetime.datetime.now().isoformat())
        print "> Creating a temporary file in the work directory..."
        print "  `-> Getting the remote user name..."
        remote_user = transport.whoami()
        print "      [remote username: {}]".format(remote_user)
        workdir = authinfo.get_workdir().format(
            username=remote_user)
        print "      [Checking/creating work directory: {}]".format(workdir)

        try:
            transport.chdir(workdir)
        except IOError:
            transport.makedirs(workdir)
            transport.chdir(workdir)

        with tempfile.NamedTemporaryFile() as f:
            fname = os.path.split(f.name)[1]
            print "  `-> Creating the file {}...".format(fname)
            remote_file_path = os.path.join(workdir, fname)
            f.write(file_content)
            f.flush()
            transport.putfile(f.name, remote_file_path)
        print "  `-> Checking if the file has been created..."
        if not transport.path_exists(remote_file_path):
            print "* ERROR! The file was not found!"
            return False
        else:
            print "      [OK]"
        print "  `-> Retrieving the file and checking its content..."

        fd, destfile = tempfile.mkstemp()
        os.close(fd)
        try:
            transport.getfile(remote_file_path, destfile)
            with open(destfile) as f:
                read_string = f.read()
            print "      [Retrieved]"
            if read_string != file_content:
                print ("* ERROR! The file content is different from what was "
                       "expected!")
                print "** Expected:"
                print file_content
                print "** Found:"
                print read_string
                return False
            else:
                print "      [Content OK]"
        finally:
            os.remove(destfile)

        print "  `-> Removing the file..."
        transport.remove(remote_file_path)
        print "  [Deleted successfully]"
        return True

    def computer_enable(self, *args):
        """
        Enable a computer.
        """
        if not is_dbenv_loaded():
            load_dbenv()

        import argparse

        from aiida.common.exceptions import NotExistent

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Enable a computer')
        # The default states are those that are shown if no option is given
        parser.add_argument('-u', '--only-for-user', type=str, metavar='EMAIL',
                            dest='user_email',
                            help="Enable a computer only for the given user. "
                                 "If not specified, enables the computer "
                                 "globally.",
                            )
        parser.add_argument('computer', type=str,
                            help="The name of the computer that you "
                                 "want to enable")

        parsed_args = parser.parse_args(args)

        user_email = parsed_args.user_email
        computername = parsed_args.computer

        try:
            computer = self.get_computer(name=computername)
        except NotExistent:
            print >> sys.stderr, "No computer exists with name '{}'".format(
                computername)
            sys.exit(1)

        if user_email is None:
            if computer.is_enabled():
                print "Computer '{}' already enabled.".format(computername)
            else:
                computer.set_enabled_state(True)
                print "Computer '{}' enabled.".format(computername)
        else:
            user_list = self.backend.users.find(email=user_email)
            if user_list is None or len(user_list) == 0:
                print >> sys.stderr, ("No user with email '{}' in the "
                                      "database.".format(user_email))
                sys.exit(1)
            user = user_list[0]
            try:
                authinfo = computer.get_authinfo(user)
                if not authinfo.enabled:
                    authinfo.enabled = True
                    print "Computer '{}' enabled for user {}.".format(
                        computername, user.get_full_name())
                else:
                    print "Computer '{}' was already enabled for user {} {}.".format(
                        computername, user.first_name, user.last_name)
            except NotExistent:
                print >> sys.stderr, ("User with email '{}' is not configured "
                                      "for computer '{}' yet.".format(
                    user_email, computername))
                sys.exit(1)

    def computer_disable(self, *args):
        """
        Disable a computer.

        If a computer is disabled, AiiDA does not try to connect to it to
        submit new calculations or check for the state of existing calculations.
        Useful, for instance, if you know that a computer is under maintenance.
        """
        if not is_dbenv_loaded():
            load_dbenv()

        import argparse

        from aiida.common.exceptions import NotExistent

        parser = argparse.ArgumentParser(
            prog=self.get_full_command_name(),
            description='Disable a computer')
        # The default states are those that are shown if no option is given
        parser.add_argument('-u', '--only-for-user', type=str, metavar='EMAIL',
                            dest='user_email',
                            help="Disable a computer only for the given user. "
                                 "If not specified, disables the computer "
                                 "globally.",
                            )
        parser.add_argument('computer', type=str,
                            help="The name of the computer that you "
                                 "want to disable")

        parsed_args = parser.parse_args(args)

        user_email = parsed_args.user_email
        computername = parsed_args.computer

        try:
            computer = self.get_computer(name=computername)
        except NotExistent:
            print >> sys.stderr, "No computer exists with name '{}'".format(
                computername)
            sys.exit(1)

        if user_email is None:
            if not computer.is_enabled():
                print "Computer '{}' already disabled.".format(computername)
            else:
                computer.set_enabled_state(False)
                print "Computer '{}' disabled.".format(computername)
        else:
            user_list = self.backend.users.find(email=user_email)
            if user_list is None or len(user_list) == 0:
                print >> sys.stderr, ("No user with email '{}' in the "
                                      "database.".format(user_email))
                sys.exit(1)
            user = user_list[0]
            try:
                authinfo = computer.get_authinfo(user)
                if authinfo.enabled:
                    authinfo.enabled = False
                    print "Computer '{}' disabled for user {}.".format(
                        computername, user.get_full_name())
                else:
                    print("Computer '{}' was already disabled for user {} {}."
                          .format(computername, user.first_name, user.last_name))
            except NotExistent:
                print >> sys.stderr, ("User with email '{}' is not configured "
                                      "for computer '{}' yet.".format(
                    user_email, computername))

    def get_computer_names(self):
        """
        Retrieve the list of computers in the DB.

        ToDo: use an API or cache the results, sometime it is quite slow!
        """
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(type='computer', project=['name'])
        if qb.count() > 0:
            return zip(*qb.all())[0]
        else:
            return None

    def get_computer(self, name):
        """
        Get a Computer object with given name, or raise NotExistent
        """
        from aiida.orm.computer import Computer as AiidaOrmComputer

        return AiidaOrmComputer.get(name)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Utilities for the workflow engine."""
import contextlib
import tornado.ioloop

from aiida.common.links import LinkType
from aiida.orm.calculation import Calculation, WorkCalculation, FunctionCalculation
from aiida.orm.data.frozendict import FrozenDict

__all__ = []

PROCESS_STATE_CHANGE_KEY = 'process|state_change|{}'
PROCESS_STATE_CHANGE_DESCRIPTION = 'The last time a process of type {}, changed state'
PROCESS_CALC_TYPES = (WorkCalculation, FunctionCalculation)


def is_work_calc_type(calc_node):
    """
    Check if the given calculation node is of the new type.
    Currently in AiiDA we have a hierarchy of 'Calculation' nodes with the following subclasses:

        1. JobCalculation
        2. InlineCalculation
        3. WorkCalculation
        4. FunctionCalculation

    1 & 2 can be considered the 'old' way of doing things, even though they are still
    in use while 3 & 4 are the 'new' way.  In loose terms the main difference is that
    the old way don't support RETURN and CALL links.

    :param calc_node: The calculation node to test
    :return: True if of the new type, False otherwise.
    """
    return isinstance(calc_node, PROCESS_CALC_TYPES)


def is_workfunction(function):
    """
    Return whether the given function is a workfunction

    :param function: a function
    :returns: True if the function is a wrapped workfunction, False otherwise
    """
    try:
        return function.is_workfunction
    except AttributeError:
        return False


def get_or_create_output_group(calculation):
    """
    For a given Calculation, get or create a new frozendict Data node that
    has as its values all output Data nodes of the Calculation.

    :param calculation: Calculation
    """
    if not isinstance(calculation, Calculation):
        raise TypeError("Can only create output groups for type Calculation")

    outputs = calculation.get_outputs_dict(link_type=LinkType.CREATE)
    outputs.update(calculation.get_outputs_dict(link_type=LinkType.RETURN))

    return FrozenDict(dict=outputs)


@contextlib.contextmanager
def loop_scope(loop):
    """
    Make an event loop current for the scope of the context

    :param loop: The event loop to make current for the duration of the scope
    :type loop: :class:`tornado.ioloop.IOLoop`
    """
    current = tornado.ioloop.IOLoop.current()

    try:
        loop.make_current()
        yield
    finally:
        current.make_current()


def set_process_state_change_timestamp(process):
    """
    Set the global setting that reflects the last time a process changed state, for the process type
    of the given process, to the current timestamp. The process type will be determined based on
    the class of the calculation node it has as its database container.

    :param process: the Process instance that changed its state
    """
    from aiida.backends.utils import set_global_setting
    from aiida.common.exceptions import UniquenessError
    from aiida.orm.calculation.inline import InlineCalculation
    from aiida.orm.calculation.job import JobCalculation
    from aiida.utils import timezone

    if isinstance(process.calc, (JobCalculation, InlineCalculation)):
        process_type = 'calculation'
    elif is_work_calc_type(process.calc):
        process_type = 'work'
    else:
        raise ValueError('unsupported calculation node type {}'.format(type(process.calc)))

    key = PROCESS_STATE_CHANGE_KEY.format(process_type)
    description = PROCESS_STATE_CHANGE_DESCRIPTION.format(process_type)
    value = timezone.now()

    try:
        set_global_setting(key, value, description)
    except UniquenessError as exception:
        process.logger.debug('could not update the {} setting because of a UniquenessError: {}'.format(key, exception))


def get_process_state_change_timestamp(process_type='calculation'):
    """
    Get the global setting that reflects the last time a process of the given process type changed
    its state. The returned value will be the corresponding timestamp or None if the setting does
    not exist.

    :param process_type: the process type for which to get the latest state change timestamp.
        Valid process types are either 'calculation' or 'work'.
    :return: a timestamp or None
    """
    from aiida.backends.utils import get_global_setting

    if process_type not in ['calculation', 'work']:
        raise ValueError('invalid value for process_type')

    key = PROCESS_STATE_CHANGE_KEY.format(process_type)

    try:
        return get_global_setting(key)
    except KeyError:
        return None



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import json

from sqlalchemy import ForeignKey
from sqlalchemy.orm import relationship
from sqlalchemy.schema import Column, UniqueConstraint
from sqlalchemy.types import Integer, Boolean, Text
from sqlalchemy.dialects.postgresql import JSONB

from aiida.transport import TransportFactory
from aiida.backends.sqlalchemy.models.base import Base
from aiida.common.exceptions import (DbContentError, MissingPluginError,
                                     ConfigurationError)
from sqlalchemy.orm.attributes import flag_modified


class DbAuthInfo(Base):
    __tablename__ = "db_dbauthinfo"

    id = Column(Integer, primary_key=True)

    aiidauser_id = Column(Integer,
                          ForeignKey('db_dbuser.id', ondelete="CASCADE", deferrable=True, initially="DEFERRED"))
    dbcomputer_id = Column(Integer,
                           ForeignKey('db_dbcomputer.id', ondelete="CASCADE", deferrable=True, initially="DEFERRED"))

    aiidauser = relationship('DbUser', backref='authinfos')
    dbcomputer = relationship('DbComputer', backref='authinfos')

    _metadata = Column('metadata', JSONB)
    auth_params = Column(JSONB)

    enabled = Column(Boolean, default=True)

    __table_args__ = (
        UniqueConstraint("aiidauser_id", "dbcomputer_id"),
    )

    def __init__(self, *args, **kwargs):
        self._metadata = dict()
        self.auth_params = dict()
        super(DbAuthInfo, self).__init__(*args, **kwargs)

    def __str__(self):
        if self.enabled:
            return "DB authorization info for {} on {}".format(self.aiidauser.email, self.dbcomputer.name)
        else:
            return "DB authorization info for {} on {} [DISABLED]".format(self.aiidauser.email, self.dbcomputer.name)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.backends.utils import set_global_setting, get_global_setting
from aiida.utils import timezone


def set_timestamp_workflow_stepper(when):
    """
    Store the current timestamp in the DbSettings for the workflow_stepper task

    :param when: can either be 'start' or 'stop', to set when the task started or stopped
    """
    task = 'workflow_stepper'

    if when == 'start':
        verb = 'started'
    elif when == 'stop':
        verb = 'finished'
    else:
        raise ValueError("the 'when' parameter can only be 'start' or 'stop'")

    set_global_setting(
        'daemon|task_{}|{}'.format(when, task), timezone.datetime.now(),
        description=(
            "The last time the daemon {} to run the task '{}'".format(verb, task)
        )
    )


def get_timestamp_workflow_stepper(when='stop'):
    """
    Return the timestamp stored in the DbSettings table for the workflow_stepper task

    :param when: can either be 'start' or 'stop', to get when the task started or stopped
    :return: a datetime.datetime object. Return None if no information is found in the DB.
    """
    task = 'workflow_stepper'

    try:
        return get_global_setting('daemon|task_{}|{}'.format(when, task))
    except KeyError:
        return None



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################



def test_and_get_code(codename, expected_code_type, use_exceptions=False):
    """
    Pass a code name and an expected code (plugin) type. Check that the
    code exists, is unique, and return the Code object. 
    
    :param codename: the name of the code to load (in the form label@machine)
    :param expected_code_type: a string with the plugin that is expected to
      be loaded. In case no plugins exist with the given name, show all existing
      plugins of that type
    :param use_exceptions: if True, raise a ValueError exception instead of
      calling sys.exit(1)
    :return: a Code object
    """
    import sys
    from aiida.common.exceptions import NotExistent
    from aiida.orm import Code

    try:
        if codename is None:
            raise ValueError
        code = Code.get_from_string(codename)
        if code.get_input_plugin_name() != expected_code_type:
            raise ValueError
    except (NotExistent, ValueError):
        from aiida.orm.querybuilder import QueryBuilder
        qb = QueryBuilder()
        qb.append(Code,
                  filters={'attributes.input_plugin':
                               {'==': expected_code_type}},
                  project='*')

        valid_code_labels = ["{}@{}".format(c.label, c.get_computer().name)
                             for [c] in qb.all()]

        if valid_code_labels:
            msg = ("Pass as further parameter a valid code label.\n"
                   "Valid labels with a {} executable are:\n".format(
                expected_code_type))
            msg += "\n".join("* {}".format(l) for l in valid_code_labels)

            if use_exceptions:
                raise ValueError(msg)
            else:
                print >> sys.stderr, msg
                sys.exit(1)
        else:
            msg = ("Code not valid, and no valid codes for {}.\n"
                   "Configure at least one first using\n"
                   "    verdi code setup".format(
                expected_code_type))
            if use_exceptions:
                raise ValueError(msg)
            else:
                print >> sys.stderr, msg
                sys.exit(1)

    return code



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
import click


def echo_info(message, bold=False):
    """
    Print an info message through click's echo function to stdout, prefixed with 'Info:'

    :param message: the string representing the message to print
    :param bold: whether to print the message in bold
    """
    click.secho('Info: ', fg='blue', bold=True, nl=False)
    click.secho(message, bold=bold)


def echo_warning(message, bold=False):
    """
    Print a warning message through click's echo function to stdout, prefixed with 'Warning:'

    :param message: the string representing the message to print
    :param bold: whether to print the message in bold
    """
    click.secho('Warning: ', fg='yellow', bold=True, nl=False)
    click.secho(message, bold=bold)


def echo_error(message, bold=False):
    """
    Print an error message through click's echo function to stdout, prefixed with 'Error:'

    :param message: the string representing the message to print
    :param bold: whether to print the message in bold
    """
    click.secho('Error: ', fg='red', bold=True, nl=False)
    click.secho(message, bold=bold)


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from abc import ABCMeta, abstractmethod, abstractproperty

from aiida.common.exceptions import UniquenessError, NotExistent, MultipleObjectsError
from aiida.common.utils import abstractclassmethod, abstractstaticmethod


def get_group_type_mapping():
    """
    Return a dictionary with ``{short_name: proper_long_name_in_DB}`` format,
    where ``short_name`` is the name to use on the command line, while
    ``proper_long_name_in_DB`` is the string stored in the ``type`` field of the
    DbGroup table.

    It is defined as a function so that the import statements are confined
    inside here.
    """
    from aiida.orm.data.upf import UPFGROUP_TYPE
    from aiida.orm.autogroup import VERDIAUTOGROUP_TYPE
    from aiida.orm.importexport import IMPORTGROUP_TYPE

    return {'data.upf': UPFGROUP_TYPE,
            'import': IMPORTGROUP_TYPE,
            'autogroup.run': VERDIAUTOGROUP_TYPE}


class AbstractGroup(object):
    """
    An AiiDA ORM implementation of group of nodes.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, **kwargs):
        """
        Create a new group. Either pass a dbgroup parameter, to reload
        ad group from the DB (and then, no further parameters are allowed),
        or pass the parameters for the Group creation.

        :param dbgroup: the dbgroup object, if you want to reload the group
            from the DB rather than creating a new one.
        :param name: The group name, required on creation
        :param description: The group description (by default, an empty string)
        :param user: The owner of the group (by default, the automatic user)
        :param type_string: a string identifying the type of group (by default,
            an empty string, indicating an user-defined group.
        """
        pass

    @abstractproperty
    def name(self):
        """
        :return: the name of the group as a string
        """
        pass

    @abstractproperty
    @name.setter
    def name(self, name):
        """
        Attempt to change the name of the group instance. If the group is already stored
        and the another group of the same type already exists with the desired name, a
        UniquenessError will be raised

        :param name: the new group name
        :raises UniquenessError: if another group of same type and name already exists
        """
        pass

    @abstractproperty
    def description(self):
        """
        :return: the description of the group as a string
        """
        pass

    @description.setter
    @abstractmethod
    def description(self, value):
        """
        :return: the description of the group as a string
        """
        pass

    @abstractproperty
    def type_string(self):
        """
        :return: the string defining the type of the group
        """
        pass

    @abstractproperty
    def user(self):
        """
        :return: a Django DbUser object, representing the user associated to
          this group.
        """
        pass

    @abstractproperty
    def dbgroup(self):
        """
        :return: the corresponding Django DbGroup object.
        """
        pass

    @abstractproperty
    def pk(self):
        """
        :return: the principal key (the ID) as an integer, or None if the
           node was not stored yet
        """
        pass

    @abstractproperty
    def id(self):
        """
        :return: the principal key (the ID) as an integer, or None if the
           node was not stored yet
        """
        pass

    @abstractproperty
    def uuid(self):
        """
        :return: a string with the uuid
        """
        pass

    @staticmethod
    def get_schema():
        """
        Every node property contains:
            - display_name: display name of the property
            - help text: short help text of the property
            - is_foreign_key: is the property foreign key to other type of the node
            - type: type of the property. e.g. str, dict, int

        :return: get schema of the group
        """
        return {
            "description": {
                "display_name": "Description",
                "help_text": "short description of the Computer",
                "is_foreign_key": False,
                "type": "str"
            },
            "id": {
                "display_name": "Id",
                "help_text": "Id of the object",
                "is_foreign_key": False,
                "type": "int"
            },
            "name": {
                "display_name": "Name",
                "help_text": "Name of the object",
                "is_foreign_key": False,
                "type": "str"
            },
            "type": {
                "display_name": "Type",
                "help_text": "Code type",
                "is_foreign_key": False,
                "type": "str"
            },
            "user_id": {
                "display_name": "Id of creator",
                "help_text": "Id of the user that created the node",
                "is_foreign_key": True,
                "related_column": "id",
                "related_resource": "_dbusers",
                "type": "int"
            },
            "uuid": {
                "display_name": "Unique ID",
                "help_text": "Universally Unique Identifier",
                "is_foreign_key": False,
                "type": "unicode"
            }
        }

    @classmethod
    def create(cls, *args, **kwargs):
        """
        Create and store a new group.
        
        Note: This method does not check for presence of the group.
        You may want to use get_or_create().

        :return: group
        """
        return cls(*args, **kwargs).store()

    @classmethod
    def get_or_create(cls, *args, **kwargs):
        """
        Try to retrieve a group from the DB with the given arguments;
        create (and store) a new group if such a group was not present yet.

        :return: (group, created) where group is the group (new or existing,
          in any case already stored) and created is a boolean saying
        """
        res = cls.query(name=kwargs.get("name"),
                        type_string=kwargs.get("type_string"))

        if res is None or len(res) == 0:
            return cls.create(*args, **kwargs), True
        elif len(res) > 1:
            raise MultipleObjectsError("More than one groups found in the "
                                       "database")
        else:
            return res[0], False

    @abstractmethod
    def __int__(self):
        """
        Convert the class to an integer. This is needed to allow querying
        with Django. Be careful, though, not to pass it to a wrong field!
        This only returns the local DB principal key (pk) value.

        :return: the integer pk of the node or None if not stored.
        """
        pass

    @abstractproperty
    def is_stored(self):
        """
        :return: True if the respective DbNode has been already saved in the
          DB, False otherwise
        """
        pass

    @abstractmethod
    def store(self):
        pass

    @abstractmethod
    def add_nodes(self, nodes):
        """
        Add a node or a set of nodes to the group.

        :note: The group must be already stored.

        :note: each of the nodes passed to add_nodes must be already stored.

        :param nodes: a Node or DbNode object to add to the group, or
          a list of Nodes or DbNodes to add.
        """
        pass

    @abstractproperty
    def nodes(self):
        """
        Return a generator/iterator that iterates over all nodes and returns
        the respective AiiDA subclasses of Node, and also allows to ask for
        the number of nodes in the group using len().
        """
        pass

    @abstractmethod
    def remove_nodes(self, nodes):
        """
        Remove a node or a set of nodes to the group.

        :note: The group must be already stored.

        :note: each of the nodes passed to add_nodes must be already stored.

        :param nodes: a Node or DbNode object to add to the group, or
          a list of Nodes or DbNodes to add.
        """
        pass

    @abstractclassmethod
    def query(cls, name=None, type_string="", pk=None, uuid=None, nodes=None,
              user=None, node_attributes=None, past_days=None, **kwargs):
        """
        Query for groups.

        :note: By default, query for user-defined groups only (type_string=="").
            If you want to query for all type of groups, pass type_string=None.
            If you want to query for a specific type of groups, pass a specific
            string as the type_string argument.

        :param name: the name of the group
        :param nodes: a node or list of nodes that belongs to the group (alternatively,
            you can also pass a DbNode or list of DbNodes)
        :param pk: the pk of the group
        :param uuid: the uuid of the group
        :param type_string: the string for the type of node; by default, look
            only for user-defined groups (see note above).
        :param user: by default, query for groups of all users; if specified,
            must be a DbUser object, or a string for the user email.
        :param past_days: by default, query for all groups; if specified, query
            the groups created in the last past_days. Must be a datetime object.
        :param node_attributes: if not None, must be a dictionary with
            format {k: v}. It will filter and return only groups where there
            is at least a node with an attribute with key=k and value=v.
            Different keys of the dictionary are joined with AND (that is, the
            group should satisfy all requirements.
            v can be a base data type (str, bool, int, float, ...)
            If it is a list or iterable, that the condition is checked so that
            there should be at least a node in the group with key=k and
            value=each of the values of the iterable.
        :param kwargs: any other filter to be passed to DbGroup.objects.filter

            Example: if ``node_attributes = {'elements': ['Ba', 'Ti'], 'md5sum': 'xxx'}``,
                it will find groups that contain the node with md5sum = 'xxx', and moreover
                contain at least one node for element 'Ba' and one node for element 'Ti'.

        """
        pass

    @classmethod
    def get(cls, *args, **kwargs):
        queryresults = cls.query(*args, **kwargs)

        if len(queryresults) == 1:
            return queryresults[0]
        elif len(queryresults) == 0:
            raise NotExistent("No Group matching the query found")
        else:
            raise MultipleObjectsError("More than one Group found -- "
                                       "I found {}".format(len(queryresults)))

    @classmethod
    def get_from_string(cls, string):
        """
        Get a group from a string.
        If only the name is provided, without colons,
        only user-defined groups are searched;
        add ':type_str' after the group name to choose also
        the type of the group equal to 'type_str'
        (e.g. 'data.upf', 'import', etc.)

        :raise ValueError: if the group type does not exist.
        :raise NotExistent: if the group is not found.
        """
        name, sep, typestr = string.rpartition(':')
        if not sep:
            name = typestr
            typestr = ""
        if typestr:
            try:
                internal_type_string = get_group_type_mapping()[typestr]
            except KeyError:
                msg = ("Invalid group type '{}'. Valid group types are: "
                       "{}".format(typestr, ",".join(sorted(
                    get_group_type_mapping().keys()))))
                raise ValueError(msg)
        else:
            internal_type_string = ""

        try:
            group = cls.get(name=name,
                            type_string=internal_type_string)
            return group
        except NotExistent:
            if typestr:
                msg = (
                    "No group of type '{}' with name '{}' "
                    "found.".format(typestr, name))
            else:
                msg = (
                    "No user-defined group with name '{}' "
                    "found.".format(name))
            raise NotExistent(msg)

    def is_user_defined(self):
        """
        :return: True if the group is user defined, False otherwise
        """
        return not self.type_string

    @abstractmethod
    def delete(self):
        """
        Delete the group from the DB
        """
        pass

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, str(self))

    def __str__(self):
        if self.type_string:
            return '"{}" [type {}], of user {}'.format(
                self.name, self.type_string, self.user.email)
        else:
            return '"{}" [user-defined], of user {}'.format(
                self.name, self.user.email)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Plugin for PBS/Torque.
This has been tested on Torque v.2.4.16 (from Ubuntu).
"""
from __future__ import division
from aiida.scheduler import Scheduler
from .pbsbaseclasses import PbsBaseClass

## These are instead the states from PBS/Torque v.2.4.16 (from Ubuntu)
#C -  Job is completed after having run [different from above, but not clashing]
#E -  Job is exiting after having run. [same as above]
#H -  Job is held. [same as above]
#Q -  job is queued, eligible to run or routed. [same as above]
#R -  job is running. [same as above]
#T -  job is being moved to new location. [same as above]
#W -  job is waiting for its execution time
#     (-a option) to be reached. [similar to above]
#S -  (Unicos only) job is suspend. [as above]


class TorqueScheduler(PbsBaseClass, Scheduler):
    """
    Subclass to support the Torque scheduler..

    I redefine only what needs to change from the base class
    """
    _logger = Scheduler._logger.getChild('torque')

    ## I don't need to change this from the base class
    #_job_resource_class = PbsJobResource

    ## For the time being I use a common dictionary, should be sufficient
    ## for the time being, but I can redefine it if needed.
    #_map_status = _map_status_pbs_common

    def _get_resource_lines(self, num_machines, num_mpiprocs_per_machine, num_cores_per_machine, max_memory_kb,
                            max_wallclock_seconds):
        """
        Return the lines for machines, memory and wallclock relative
        to pbspro.
        """
        return_lines = []

        select_string = "nodes={}".format(num_machines)
        if num_cores_per_machine:
            select_string += ":ppn={}".format(num_cores_per_machine)
        elif num_mpiprocs_per_machine:
            # if num_cores_per_machine is not defined then use
            # num_mpiprocs_per_machine
            select_string += ":ppn={}".format(num_mpiprocs_per_machine)

        if max_wallclock_seconds is not None:
            try:
                tot_secs = int(max_wallclock_seconds)
                if tot_secs <= 0:
                    raise ValueError
            except ValueError:
                raise ValueError("max_wallclock_seconds must be "
                                 "a positive integer (in seconds)! It is instead '{}'"
                                 "".format(max_wallclock_seconds))
            hours = tot_secs // 3600
            tot_minutes = tot_secs % 3600
            minutes = tot_minutes // 60
            seconds = tot_minutes % 60
            # There is always something before, at least the total #
            # of nodes
            select_string += (",walltime={:02d}:{:02d}:{:02d}".format(hours, minutes, seconds))

        if max_memory_kb:
            try:
                virtualMemoryKb = int(max_memory_kb)
                if virtualMemoryKb <= 0:
                    raise ValueError
            except ValueError:
                raise ValueError("max_memory_kb must be "
                                 "a positive integer (in kB)! It is instead '{}'"
                                 "".format((max_memory_kb)))
            # There is always something before, at least the total #
            # of nodes
            select_string += ",mem={}kb".format(virtualMemoryKb)

        return_lines.append("#PBS -l {}".format(select_string))
        return return_lines



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import functools
import os
import shutil

from sqlalchemy.orm import sessionmaker

import aiida.backends.sqlalchemy
from aiida.backends.settings import AIIDADB_PROFILE
from aiida.backends.sqlalchemy.models.base import Base
from aiida.backends.sqlalchemy.models.computer import DbComputer
from aiida.backends.sqlalchemy.models.user import DbUser
from aiida.backends.sqlalchemy.utils import install_tc
from aiida.backends.testimplbase import AiidaTestImplementation
from aiida.common.setup import get_profile_config
from aiida.common.utils import get_configured_user_email
from aiida.orm.computer import Computer


# Querying for expired objects automatically doesn't seem to work.
# That's why expire on commit=False resolves many issues of objects beeing
# obsolete

expire_on_commit = True
Session = sessionmaker(expire_on_commit=expire_on_commit)


# This contains the codebase for the setUpClass and tearDown methods used
# internally by the AiidaTestCase. This inherits only from 'object' to avoid
# that it is picked up by the automatic discovery of tests
# (It shouldn't, as it risks to destroy the DB if there are not the checks
# in place, and these are implemented in the AiidaTestCase
class SqlAlchemyTests(AiidaTestImplementation):

    # Specify the need to drop the table at the beginning of a test case
    # If True, completely drops the tables and recreates the schema, 
    # but this is usually unnecessary and pretty slow
    # Also, if the tests are interrupted, there is the risk that the
    # DB remains dropped, so you have to do 'verdi -p test_xxx setup' again to
    # install the schema again
    drop_all = False

    test_session = None
    connection = None

    def setUpClass_method(self):

        from aiida.backends.sqlalchemy import get_scoped_session

        if self.test_session is None:
            # Should we use reset_session?
            self.test_session = get_scoped_session()

        if self.drop_all:
            Base.metadata.drop_all(self.test_session.connection)
            Base.metadata.create_all(self.test_session.connection)
            install_tc(self.test_session.connection)
        else:
            self.clean_db()

        self.insert_data()

    def setUp_method(self):
        pass

    def tearDown_method(self):
        pass

    def insert_data(self):
        """
        Insert default data into the DB.
        """
        email = get_configured_user_email()

        has_user = DbUser.query.filter(DbUser.email==email).first()
        if not has_user:
            self.user = DbUser(get_configured_user_email(), "foo", "bar",
                               "tests")
            self.test_session.add(self.user)
            self.test_session.commit()
        else:
            self.user = has_user

        # Required by the calling class
        self.user_email = self.user.email

        # Also self.computer is required by the calling class
        has_computer = DbComputer.query.filter(DbComputer.hostname ==
                                               'localhost').first()
        if not has_computer:
            self.computer = SqlAlchemyTests._create_computer()
            self.computer.store()
        else:
            self.computer = Computer(dbcomputer=has_computer)

    @staticmethod
    def _create_computer(**kwargs):
        defaults = dict(name='localhost',
                        hostname='localhost',
                        transport_type='local',
                        scheduler_type='pbspro',
                        workdir='/tmp/aiida')
        defaults.update(kwargs)
        return Computer(**defaults)

    @staticmethod
    def inject_computer(f):
        @functools.wraps(f)
        def dec(*args, **kwargs):
            computer = DbComputer.query.filter_by(name="localhost").first()
            args = list(args)
            args.insert(1, computer)
            return f(*args, **kwargs)

        return dec

    def clean_db(self):
        from aiida.backends.sqlalchemy.models.computer import DbComputer
        from aiida.backends.sqlalchemy.models.workflow import DbWorkflow, table_workflowstep_calc, \
            table_workflowstep_subworkflow, DbWorkflowStep, DbWorkflowData
        from aiida.backends.sqlalchemy.models.group import DbGroup
        from aiida.backends.sqlalchemy.models.node import DbLink
        from aiida.backends.sqlalchemy.models.node import DbNode
        from aiida.backends.sqlalchemy.models.log import DbLog
        from aiida.backends.sqlalchemy.models.user import DbUser

        # Delete the workflows
        # Complicated way to make sure we 'unwind' all the relationships
        # between workflows and their children.
        self.test_session.connection().execute(table_workflowstep_calc.delete())
        self.test_session.connection().execute(table_workflowstep_subworkflow.delete())
        self.test_session.query(DbWorkflowData).delete()
        self.test_session.query(DbWorkflowStep).delete()
        self.test_session.query(DbWorkflow).delete()

        # Empty the relationship dbgroup.dbnode
        dbgroups = self.test_session.query(DbGroup).all()
        for dbgroup in dbgroups:
            dbgroup.dbnodes = []

        # Delete the groups
        self.test_session.query(DbGroup).delete()

        # I first need to delete the links, because in principle I could
        # not delete input nodes, only outputs. For simplicity, since
        # I am deleting everything, I delete the links first
        self.test_session.query(DbLink).delete()

        # Then I delete the nodes, otherwise I cannot
        # delete computers and users
        self.test_session.query(DbNode).delete()

        # # Delete the users
        self.test_session.query(DbUser).delete()

        # Delete the computers
        self.test_session.query(DbComputer).delete()

        # Delete the logs
        self.test_session.query(DbLog).delete()

        self.test_session.commit()

    def tearDownClass_method(self):
        from aiida.settings import REPOSITORY_PATH
        from aiida.common.setup import TEST_KEYWORD
        from aiida.common.exceptions import InvalidOperation
        if TEST_KEYWORD not in REPOSITORY_PATH:
            raise InvalidOperation("Be careful. The repository for the tests "
                                   "is not a test repository. I will not "
                                   "empty the database and I will not delete "
                                   "the repository. Repository path: "
                                   "{}".format(REPOSITORY_PATH))

        self.clean_db()

        self.test_session.close()
        self.test_session = None

        # I clean the test repository
        shutil.rmtree(REPOSITORY_PATH, ignore_errors=True)
        os.makedirs(REPOSITORY_PATH)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

import plumpy

from aiida.backends.testbase import AiidaTestCase
from aiida import work
from aiida.workflows.wf_demo import WorkflowDemo
from aiida.daemon.workflowmanager import execute_steps
from . import utils


class Proc(work.Process):
    def _run(self):
        pass


def the_hans_klok_comeback(loop):
    loop.stop()


class TestWorkchain(AiidaTestCase):
    def setUp(self):
        super(TestWorkchain, self).setUp()
        self.runner = utils.create_test_runner()

    def tearDown(self):
        self.runner.close()
        self._runner = None

    def test_call_on_calculation_finish(self):
        loop = self.runner.loop
        proc = Proc(runner=self.runner)
        future = plumpy.Future()

        def calc_done(pk):
            self.assertEqual(pk, proc.calc.pk)
            loop.stop()
            future.set_result(True)

        self.runner.call_on_calculation_finish(proc.calc.pk, calc_done)

        # Run the calculation
        self.runner.loop.add_callback(proc.step_until_terminated)
        self._run_loop_for(5.)

        self.assertTrue(future.result())

    def test_call_on_wf_finish(self):
        loop = self.runner.loop
        future = plumpy.Future()

        # Need to start() so it's stored
        wf = WorkflowDemo()
        wf.start()

        def wf_done(pk):
            self.assertEqual(pk, wf.pk)
            loop.stop()
            future.set_result(True)

        self.runner.call_on_legacy_workflow_finish(wf.pk, wf_done)

        # Run the wf
        while wf.is_running():
            execute_steps()

        self._run_loop_for(10.)
        self.assertTrue(future.result())

    def _run_loop_for(self, seconds):
        loop = self.runner.loop
        loop.call_later(seconds, the_hans_klok_comeback, self.runner.loop)
        loop.start()


class TestRunner(AiidaTestCase):
    def setUp(self):
        super(TestWorkchain, self).setUp()
        self.runner = work.runners.new_runner(rmq_submit=True)

    def tearDown(self):
        super(TestWorkchain, self).tearDown()
        self.runner.close()
        self.runner = None
        work.runners.set_runner(None)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm import CalculationFactory
from aiida.parsers.parser import Parser
from aiida.orm.data.parameter import ParameterData

TemplatereplacerCalculation = CalculationFactory('simpleplugins.templatereplacer')

class TemplatereplacerDoublerParser(Parser):

    def __init__(self, calc):
        """
        Initialize the Parser for a TemplatereplacerCalculation

        :param calculation: instance of the TemplatereplacerCalculation
        """
        if not isinstance(calc, TemplatereplacerCalculation):
            raise ValueError('Input calculation must be of type {}'.format(type(TemplatereplacerCalculation)))

        super(TemplatereplacerDoublerParser, self).__init__(calc)

    def parse_with_retrieved(self, retrieved):
        """
        Parse the output nodes for a PwCalculations from a dictionary of retrieved nodes.
        Two nodes that are expected are the default 'retrieved' FolderData node which will
        store the retrieved files permanently in the repository. The second required node
        is the unstored FolderData node with the temporary retrieved files, which should
        be passed under the key 'retrieved_temporary_folder_key' of the Parser class.

        :param retrieved: a dictionary of retrieved nodes
        """
        import os

        output_nodes = []

        try:
            output_file = self._calc.inp.template.get_dict()['output_file_name']
        except KeyError:
            self.logger.error("the output file name 'output_file_name' was not specified in the 'template' input node")
            return False, ()

        retrieved_folder = retrieved[self._calc._get_linkname_retrieved()]
        try:
            parsed_value = int(retrieved_folder.get_file_content(output_file).strip())
        except (AttributeError, IOError, ValueError) as e:
            self.logger.error("* UNABLE TO RETRIEVE VALUE for calc pk={}: I got {}: {}".format(self._calc.pk, type(e), e))
            return False, ()

        output_dict = {
            'value': parsed_value,
            'retrieved_temporary_files': []
        }

        try:
            retrieve_temporary_files = self._calc.inp.template.get_dict()['retrieve_temporary_files']
        except KeyError:
            retrieve_temporary_files = None

        # If the 'retrieve_temporary_files' key was set in the template input node, we expect a temporary
        # FolderData node in the 'retrieved' arguments
        if retrieve_temporary_files is not None:
            try:
                retrieved_temporary_folder = retrieved[self.retrieved_temporary_folder_key]
            except KeyError:
                self.logger.error('the {} was not passed as an argument'.format(self.retrieved_temporary_folder_key))
                return False, ()

            for retrieved_file in retrieve_temporary_files:

                file_path = os.path.join(retrieved_temporary_folder, retrieved_file)

                if not os.path.isfile(file_path):
                    self.logger.error('the file {} was not found in the temporary retrieved folder {}'
                        .format(retrieved_file, retrieved_temporary_folder))
                    return False, ()

                with open(file_path, 'r') as handle:
                    parsed_value = handle.read().strip()

                # We always strip the content of the file from whitespace to simplify testing for expected output
                output_dict['retrieved_temporary_files'].append((retrieved_file, parsed_value))

        output_parameters = ParameterData(dict=output_dict)
        output_nodes.append((self.get_linkname_outparams(), output_parameters))

        return True, output_nodes


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.work.processes import Process



class DummyProcess(Process):
    """
    A Process that does nothing when it runs.
    """
    @classmethod
    def define(cls, spec):
        super(DummyProcess, cls).define(spec)
        spec.dynamic_input()
        spec.dynamic_output()

    def _run(self):
        pass


class BadOutput(Process):
    """
    A Process that emits an output that isn't part of the spec raising an
    exception.
    """
    @classmethod
    def define(cls, spec):
        super(BadOutput, cls).define(spec)
        spec.dynamic_output()

    def _run(self):
        self.out("bad_output", 5)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.orm.implementation.calculation import Calculation
from aiida.orm.mixins import FunctionCalculationMixin


class FunctionCalculation(FunctionCalculationMixin, Calculation):
    """
    Calculation node to record the results of a function that was run wrapped by the
    :class:`aiida.work.workfunctions.workfunction` decorator
    """


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.backends.testbase import AiidaTestCase


class TestQueryBuilderDjango(AiidaTestCase):

    def test_clsf_django(self):
        """
        This tests the classifications of the QueryBuilder u. the django backend.
        """
        from aiida.backends.djsite.querybuilder_django.dummy_model import (
            DbNode, DbUser, DbComputer,
            DbGroup,
        )
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.data.structure import StructureData
        from aiida.orm import Group, Node, Computer, Data, Calculation
        from aiida.common.exceptions import InputValidationError
        qb = QueryBuilder()

        with self.assertRaises(InputValidationError):
            qb._get_ormclass(None, 'data')
        with self.assertRaises(InputValidationError):
            qb._get_ormclass(None, 'data.Data')
        with self.assertRaises(InputValidationError):
            qb._get_ormclass(None, '.')

        for cls, clstype, query_type_string in (
                qb._get_ormclass(StructureData, None),
                qb._get_ormclass(None, 'data.structure.StructureData.'),
        ):
            self.assertEqual(clstype, 'data.structure.StructureData.')
            self.assertTrue(issubclass(cls, DbNode))
            self.assertEqual(clstype, 'data.structure.StructureData.')
            self.assertEqual(query_type_string,
                             StructureData._query_type_string)

        for cls, clstype, query_type_string in (
                qb._get_ormclass(Node, None),
                qb._get_ormclass(DbNode, None),
                qb._get_ormclass(None, '')
        ):
            self.assertEqual(clstype, Node._plugin_type_string)
            self.assertEqual(query_type_string, Node._query_type_string)
            self.assertTrue(issubclass(cls, DbNode))

        for cls, clstype, query_type_string in (
                qb._get_ormclass(DbGroup, None),
                qb._get_ormclass(Group, None),
                qb._get_ormclass(None, 'group'),
                qb._get_ormclass(None, 'Group'),
        ):
            self.assertEqual(clstype, 'group')
            self.assertEqual(query_type_string, None)
            self.assertTrue(issubclass(cls, DbGroup))

        for cls, clstype, query_type_string in (
                qb._get_ormclass(DbUser, None),
                qb._get_ormclass(DbUser, None),
                qb._get_ormclass(None, "user"),
                qb._get_ormclass(None, "User"),
        ):
            self.assertEqual(clstype, 'user')
            self.assertEqual(query_type_string, None)
            self.assertTrue(issubclass(cls, DbUser))

        for cls, clstype, query_type_string in (
                qb._get_ormclass(DbComputer, None),
                qb._get_ormclass(Computer, None),
                qb._get_ormclass(None, 'computer'),
                qb._get_ormclass(None, 'Computer'),
        ):
            self.assertEqual(clstype, 'computer')
            self.assertEqual(query_type_string, None)
            self.assertTrue(issubclass(cls, DbComputer))

        for cls, clstype, query_type_string in (
                qb._get_ormclass(Data, None),
                qb._get_ormclass(None, 'data.Data.'),
        ):
            self.assertEqual(clstype, Data._plugin_type_string)
            self.assertEqual(query_type_string, Data._query_type_string)
            self.assertTrue(issubclass(cls, DbNode))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################

from aiida.backends.testbase import AiidaTestCase
from aiida.backends.utils import get_workflow_list
from aiida.common.datastructures import wf_states
from aiida.orm.backend import construct_backend
from aiida.workflows.test import WFTestEmpty
from aiida.orm.implementation import get_workflow_info
from aiida.workflows.test import WFTestSimpleWithSubWF


class TestWorkflowBasic(AiidaTestCase):
    """
    These tests check the basic features of workflows.
    Now only the load_workflow function is tested.
    """

    def test_load_workflows(self):
        """
        Test for load_node() function.
        """
        from aiida.orm import load_workflow

        a = WFTestEmpty()
        a.store()

        self.assertEquals(a.pk, load_workflow(wf_id=a.pk).pk)
        self.assertEquals(a.pk, load_workflow(wf_id=a.uuid).pk)
        self.assertEquals(a.pk, load_workflow(pk=a.pk).pk)
        self.assertEquals(a.pk, load_workflow(uuid=a.uuid).pk)

        with self.assertRaises(ValueError):
            load_workflow(wf_id=a.pk, pk=a.pk)
        with self.assertRaises(ValueError):
            load_workflow(pk=a.pk, uuid=a.uuid)
        with self.assertRaises(ValueError):
            load_workflow(pk=a.uuid)
        with self.assertRaises(ValueError):
            load_workflow(uuid=a.pk)
        with self.assertRaises(ValueError):
            load_workflow()

    def test_listing_workflows(self):
        """
        Test ensuring that the workflow listing works as expected.
        (Listing initialized & running workflows and not listing finished
        workflows or workflows with errors).
        """
        backend = construct_backend()

        # Assuming there is only one user
        user = backend.users.find(email=self.user_email)[0]
        # Creating a workflow & storing it
        a = WFTestEmpty()
        a.store()

        # Setting manually the state to RUNNING.
        a.set_state(wf_states.RUNNING)

        # Getting all the available workflows of the current user
        # and checking if we got the right one.
        wfqs = get_workflow_list(all_states=True, user=user)
        self.assertTrue(len(wfqs) == 1, "We expect one workflow")
        a_prime = wfqs[0].get_aiida_class()
        self.assertEqual(a.uuid, a_prime.uuid, "The uuid is not the expected "
                                               "one")

        # We ask all the running workflows. We should get one workflow.
        wfqs = get_workflow_list(all_states=True, user=user)
        self.assertTrue(len(wfqs) == 1, "We expect one workflow")
        a_prime = wfqs[0].get_aiida_class()
        self.assertEqual(a.uuid, a_prime.uuid, "The uuid is not the expected "
                                               "one")

        # We change the state of the workflow to FINISHED.
        a.set_state(wf_states.FINISHED)

        # Getting all the available workflows of the current user
        # and checking if we got the right one.
        wfqs = get_workflow_list(all_states=True, user=user)
        self.assertTrue(len(wfqs) == 1, "We expect one workflow")
        a_prime = wfqs[0].get_aiida_class()
        self.assertEqual(a.uuid, a_prime.uuid, "The uuid is not the expected "
                                               "one")

        # We ask all the running workflows. We should get zero results.
        wfqs = get_workflow_list(all_states=False, user=user)
        self.assertTrue(len(wfqs) == 0, "We expect zero workflows")

        # We change the state of the workflow to INITIALIZED.
        a.set_state(wf_states.INITIALIZED)

        # We ask all the running workflows. We should get one workflow.
        wfqs = get_workflow_list(all_states=True, user=user)
        self.assertTrue(len(wfqs) == 1, "We expect one workflow")
        a_prime = wfqs[0].get_aiida_class()
        self.assertEqual(a.uuid, a_prime.uuid, "The uuid is not the expected "
                                               "one")

        # We change the state of the workflow to ERROR.
        a.set_state(wf_states.ERROR)

        # We ask all the running workflows. We should get zero results.
        wfqs = get_workflow_list(all_states=False, user=user)
        self.assertTrue(len(wfqs) == 0, "We expect zero workflows")

    def test_workflow_info(self):
        """
        This test checks that the workflow info is generate without any
        exceptions
        :return:
        """
        backend = construct_backend()

        # Assuming there is only one user
        user = backend.users.find(email=self.user_email)[0]

        # Creating a simple workflow & storing it
        a = WFTestEmpty()
        a.store()

        # Emulate the workflow list
        for w in get_workflow_list(all_states=True, user=user):
            if not w.is_subworkflow():
                get_workflow_info(w)

        # Create a workflow with sub-workflows and store it
        b = WFTestSimpleWithSubWF()
        b.store()

        # Emulate the workflow list
        for w in get_workflow_list(all_states=True, user=user):
            if not w.is_subworkflow():
                get_workflow_info(w)

        # Start the first workflow and perform a workflow list
        b.start()
        for w in get_workflow_list(all_states=True, user=user):
            if not w.is_subworkflow():
                get_workflow_info(w)

    def test_wf_get_state(self):
        """
        Simple test that checks the state of the workflows. We create two
        workflows since the test order in the SQLA was influencing the value
        of aiida.backends.sqlalchemy.models.workflow.DbWorkflow.state which
        should be a Choice object, according to the SQLA doc. Sometimes it
        was automatically converted to unicode.

        Since we are interested to get a unicode from
        aiida.orm.implementation.general.workflow.AbstractWorkflow#get_state
        we enforce this conversion at
        aiida.orm.implementation.sqlalchemy.workflow.Workflow#get_state

        For more info, check issue #951
        """
        # Creating two simple workflows & storing them
        wf1 = WFTestEmpty()
        wf1.store()

        wf2 = WFTestEmpty()
        wf2.store()

        # Checking that the get_state doesn't throw exceptions and that
        # it is a valid state
        self.assertIn(wf1.get_state(), wf_states)
        self.assertIn(wf2.get_state(), wf_states)

    def test_wf_ctime(self):
        import datetime
        import pytz

        # Get the current datetime (before the creation of the workflow)
        dt_before = datetime.datetime.now(pytz.utc)

        # Creating a simple workflow & storing it
        wf = WFTestEmpty()
        wf.store()

        # Get the current datetime (after the creation of the workflow)
        dt_after = datetime.datetime.now(pytz.utc)

        self.assertLessEqual(dt_before, wf.ctime, "The workflow doesn't have"
                                                  "a valid creation time")

        self.assertGreaterEqual(dt_after, wf.ctime, "The workflow doesn't "
                                                    "have a valid creation "
                                                    "time")

    def test_failing_calc_in_wf(self):
        """
        This test checks that a workflow (but also a workflow with
        sub-workflows) that has an exception at one of its steps stops
        properly and it is not left as RUNNING.
        """
        from aiida.daemon.workflowmanager import execute_steps
        from aiida.workflows.test import (FailingWFTestSimple,
                                          FailingWFTestSimpleWithSubWF)

        try:
            # Testing the error propagation of a simple workflow
            wf = FailingWFTestSimple()
            wf.store()
            step_no = 0
            wf.start()
            while wf.is_running():
                execute_steps()
                step_no += 1
                self.assertLess(step_no, 5, "This workflow should have stopped "
                                            "since it is failing")

            # Testing the error propagation of a workflow with subworkflows
            wf = FailingWFTestSimpleWithSubWF()
            wf.store()

            step_no = 0
            wf.start()
            while wf.is_running():
                execute_steps()
                step_no += 1
                self.assertLess(step_no, 5, "This workflow should have stopped "
                                            "since it is failing")
        finally:
            pass

    def test_result_parameter_name_colision(self):
        """
        This test checks that the the workflow parameters and results do not
        collide. This was a problem in SQLA (Issue #960) but a test for both
        backends is added (for completeness).
        """
        # Creating a simple workflow & storing it
        wf = WFTestEmpty()
        wf.store()

        # Set some parameters
        params = {'band_calculation_set': 2,
                  'codename': 'pw-5.2.0',
                  'pseudo_family': 'SSSP_v0.7_eff_PBE'}
        wf.set_params(params)

        # Add some results that their names collide with the parameter names
        wf.add_result('structure', 'test_string_1')
        wf.add_result('codename', 'test_string_2')

        # Check that we have the correct results
        self.assertDictEqual(
            {'structure': 'test_string_1', 'codename': 'test_string_2'},
            wf.get_results(), "The workflow results are not the expected "
                              "ones.")

        # Check that we have the correct parameters
        self.assertDictEqual(params, wf.get_parameters(),
                             "The workflow parameters are not the expected "
                             "ones.")

    def tearDown(self):
        """
        Cleaning the database after each test. Since I don't
        want the workflows of one test to interfere with the
        workflows of the other tests.
        """
        self._class_was_setup = True
        self.clean_db()
        self.insert_data()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm import Data
import os


class RemoteData(Data):
    """
    Store a link to a file or folder on a remote machine.
    
    Remember to pass a computer!
    """

    def get_dbcomputer(self):
        return self.dbnode.dbcomputer

    def get_computer_name(self):
        return self.get_computer().name

    def get_remote_path(self):
        return self.get_attr('remote_path')

    def set_remote_path(self, val):
        self._set_attr('remote_path', val)

    def add_path(self, src_abs, dst_filename=None):
        """
        Disable adding files or directories to a RemoteData
        """
        from aiida.common.exceptions import ModificationNotAllowed

        raise ModificationNotAllowed("Cannot add files or directories to a RemoteData object")

    def is_empty(self):
        """
        Check if remote folder is empty
        """
        authinfo = self._get_authinfo()
        t = authinfo.get_transport()

        with t:
            try:
                t.chdir(self.get_remote_path())
            except IOError as e:
                if e.errno == 2:  # directory not existing
                    return True  # is indeed empty, i.e. unusable
            return not t.listdir()

    def getfile(self, relpath, destpath):
        """
        Connects to the remote folder and gets a string with the (full) content of the file.

        :param relpath: The relative path of the file to show.
        :param destpath: A path on the local computer to get the file
        :return: a string with the file content
        """
        authinfo = self._get_authinfo()
        t = authinfo.get_transport()

        with t:
            try:
                full_path = os.path.join(self.get_remote_path(), relpath)
                t.getfile(full_path, destpath)
            except IOError as e:
                if e.errno == 2:  # file not existing
                    raise IOError("The required remote file {} on {} does not exist or has been deleted.".format(
                        full_path, self.get_computer().name
                    ))
                else:
                    raise

            return t.listdir()

    def listdir(self, relpath="."):
        """
        Connects to the remote folder and lists the directory content.

        :param relpath: If 'relpath' is specified, lists the content of the given subfolder.
        :return: a flat list of file/directory names (as strings).
        """
        authinfo = self._get_authinfo()
        t = authinfo.get_transport()

        with t:
            try:
                full_path = os.path.join(self.get_remote_path(), relpath)
                t.chdir(full_path)
            except IOError as e:
                if e.errno == 2 or e.errno == 20:  # directory not existing or not a directory
                    exc = IOError(
                        "The required remote folder {} on {} does not exist, is not a directory or has been deleted.".format(
                            full_path, self.get_computer().name
                        ))
                    exc.errno = e.errno
                    raise exc
                else:
                    raise

            try:
                return t.listdir()
            except IOError as e:
                if e.errno == 2 or e.errno == 20:  # directory not existing or not a directory
                    exc = IOError(
                        "The required remote folder {} on {} does not exist, is not a directory or has been deleted.".format(
                            full_path, self.get_computer().name
                        ))
                    exc.errno = e.errno
                    raise exc
                else:
                    raise

    def listdir_withattributes(self, path="."):
        """
        Connects to the remote folder and lists the directory content.

        :param relpath: If 'relpath' is specified, lists the content of the given subfolder.
        :return: a list of dictionaries, where the documentation is in :py:class:Transport.listdir_withattributes.
        """
        authinfo = self._get_authinfo()
        t = authinfo.get_transport()

        with t:
            try:
                full_path = os.path.join(self.get_remote_path(), path)
                t.chdir(full_path)
            except IOError as e:
                if e.errno == 2 or e.errno == 20:  # directory not existing or not a directory
                    exc = IOError(
                        "The required remote folder {} on {} does not exist, is not a directory or has been deleted.".format(
                            full_path, self.get_computer().name
                        ))
                    exc.errno = e.errno
                    raise exc
                else:
                    raise

            try:
                return t.listdir_withattributes()
            except IOError as e:
                if e.errno == 2 or e.errno == 20:  # directory not existing or not a directory
                    exc = IOError(
                        "The required remote folder {} on {} does not exist, is not a directory or has been deleted.".format(
                            full_path, self.get_computer().name
                        ))
                    exc.errno = e.errno
                    raise exc
                else:
                    raise

    def _clean(self):
        """
        Remove all content of the remote folder on the remote computer
        """
        import os

        authinfo = self._get_authinfo()
        t = authinfo.get_transport()

        remote_dir = self.get_remote_path()
        pre, post = os.path.split(remote_dir)

        with t:
            try:
                t.chdir(pre)
                t.rmtree(post)
            except IOError as e:
                if e.errno == 2:  # directory not existing
                    pass

    def _validate(self):
        from aiida.common.exceptions import ValidationError

        super(RemoteData, self)._validate()

        try:
            self.get_remote_path()
        except AttributeError:
            raise ValidationError("attribute 'remote_path' not set.")

        computer = self.get_computer()
        if computer is None:
            raise ValidationError("Remote computer not set.")

    def _get_authinfo(self):
        return self.backend.authinfos.get(self.get_computer(), self.get_user())



#!/usr/bin/env runaiida
# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.orm.data.base import NumericType
from aiida.orm.data.float import Float
from aiida.orm.data.int import Int
from aiida import work

"""
This example illustrates in a very minimal way how a WorkChain can be defined
and how it can be run. This mostly illustrates how the spec of the WorkChain
is defined and how functions in the outline of the spec have to be defined.
"""


class SumWorkChain(work.WorkChain):

    @classmethod
    def define(cls, spec):
        super(SumWorkChain, cls).define(spec)
        spec.input('a', valid_type=NumericType)
        spec.input('b', valid_type=NumericType)
        spec.outline(
            cls.sum
        )
        spec.output('sum', valid_type=NumericType)

    def sum(self):
        self.out('sum', self.inputs.a + self.inputs.b)


class ProductWorkChain(work.WorkChain):

    @classmethod
    def define(cls, spec):
        super(ProductWorkChain, cls).define(spec)
        spec.input('a', valid_type=NumericType)
        spec.input('b', valid_type=NumericType)
        spec.outline(
            cls.product
        )
        spec.output('product', valid_type=NumericType)

    def product(self):
        self.out('product', self.inputs.a * self.inputs.b)
    

class SumProductWorkChain(work.WorkChain):

    @classmethod
    def define(cls, spec):
        super(SumProductWorkChain, cls).define(spec)
        spec.input('a', valid_type=NumericType)
        spec.input('b', valid_type=NumericType)
        spec.input('c', valid_type=NumericType)
        spec.outline(
            cls.sum,
            cls.product
        )
        spec.output('sumproduct', valid_type=NumericType)

    def sum(self):
        self.ctx.sum = self.inputs.a + self.inputs.b

    def product(self):
        self.out('sumproduct', self.ctx.sum * self.inputs.c)


def main():
    inputs = {
        'a': Float(3.14),
        'b': Int(4),
        'c': Int(6)
    }

    results = work.run(SumWorkChain, **inputs)
    print 'Result of SumWorkChain: {}'.format(results)

    results = work.run(ProductWorkChain, **inputs)
    print 'Result of ProductWorkChain: {}'.format(results)

    results = work.run(SumProductWorkChain, **inputs)
    print 'Result of SumProductWorkChain: {}'.format(results)


if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida.restapi.translator.data import DataTranslator
from aiida.restapi.translator.node import NodeTranslator
from aiida.restapi.common.exceptions import RestInputValidationError

class UpfDataTranslator(DataTranslator):
    """
    Translator relative to resource 'upfs' and aiida class UpfData
    """

    # A label associated to the present class (coincides with the resource name)
    __label__ = "upfs"
    # The AiiDA class one-to-one associated to the present class
    from aiida.orm.data.upf import UpfData
    _aiida_class = UpfData
    # The string name of the AiiDA class
    _aiida_type = "data.upf.UpfData"
    # The string associated to the AiiDA class in the query builder lexicon
    _qb_type = _aiida_type + '.'

    _result_type = __label__

    def __init__(self, **kwargs):
        """
        Initialise the parameters.
        Create the basic query_help
        """
        super(UpfDataTranslator, self).__init__(Class=self.__class__,
                                                **kwargs)

    @staticmethod
    def get_visualization_data(node, format=None):
        """

        Returns: data in a format required by dr.js to visualize a 2D plot
        with multiple data series.

        """
        return []

    @staticmethod
    def get_downloadable_data(node, format=None):
        """
        Generic function extented for kpoints data. Currently
        it is not implemented.

        :param node: node object that has to be visualized
        :param format: file extension format
        :returns: raise RestFeatureNotAvailable exception
        """

        response = {}

        if node.folder.exists():
            folder_node = node._get_folder_pathsubfolder
            filename = node.filename

            try:
                content = NodeTranslator.get_file_content(folder_node, filename)
            except IOError as e:
                error = "Error in getting {} content".format(filename)
                raise RestInputValidationError (error)

            response["status"] = 200
            response["data"] = content
            response["filename"] = filename

        else:
            response["status"] = 200
            response["data"] = "file does not exist"

        return response





# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=missing-docstring,invalid-name

import mock
from click.testing import CliRunner

from aiida.backends.testbase import AiidaTestCase
from aiida.common.datastructures import calc_states
from aiida.utils.capturing import Capturing
from aiida.work import WorkChain

# Common computer information
computer_common_info = [
    "localhost",
    "",
    "True",
    "ssh",
    "torque",
    "#!/bin/bash",
    "/scratch/{username}/aiida_run",
    "mpirun -np {tot_num_mpiprocs}",
    "1",
    EOFError,
    EOFError,
]

# Computer #1
computer_name_1 = "torquessh1"
computer_setup_input_1 = [computer_name_1] + computer_common_info

# Computer #2
computer_name_2 = "torquessh2"
computer_setup_input_2 = [computer_name_2] + computer_common_info

# Common code information
code_common_info_1 = [
    "simple script",
    "False",
    "simpleplugins.templatereplacer",
]
code_common_info_2 = [
    "/usr/local/bin/doubler.sh",
    EOFError,
    EOFError,
]

# Code #1
code_name_1 = "doubler_1"
code_setup_input_1 = ([code_name_1] + code_common_info_1 + [computer_name_1] + code_common_info_2)
# Code #2
code_name_2 = "doubler_2"
code_setup_input_2 = ([code_name_2] + code_common_info_1 + [computer_name_2] + code_common_info_2)

# User #1
user_1 = {
    'email': "testuser1@localhost",
    'first_name': "Max",
    'last_name': "Mueller",
    'institution': "Testing Instiute"
}
# User #2
user_2 = {
    'email': "testuser2@localhost",
    'first_name': "Sabine",
    'last_name': "Mueller",
    'institution': "Testing Instiute"
}


# pylint: disable=protected-access
class TestVerdiCalculationCommands(AiidaTestCase):

    # pylint: disable=no-member
    # pylint says: Method 'computer' has no 'name' / 'id' member
    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create some calculations with various states
        """
        super(TestVerdiCalculationCommands, cls).setUpClass()

        from aiida.orm import JobCalculation

        # Create some calculation
        calc1 = JobCalculation(
            computer=cls.computer, resources={
                'num_machines': 1,
                'num_mpiprocs_per_machine': 1
            }).store()
        calc1._set_state(calc_states.TOSUBMIT)
        calc2 = JobCalculation(
            computer=cls.computer.name, resources={
                'num_machines': 1,
                'num_mpiprocs_per_machine': 1
            }).store()
        calc2._set_state(calc_states.COMPUTED)
        calc3 = JobCalculation(
            computer=cls.computer.id, resources={
                'num_machines': 1,
                'num_mpiprocs_per_machine': 1
            }).store()
        calc3._set_state(calc_states.FINISHED)

    def test_calculation_list(self):
        """
        Do some calculation listing to ensure that verdi calculation list
        works and gives at least to some extent the expected results.
        """
        from aiida.cmdline.commands.calculation import Calculation
        calc_cmd = Calculation()

        with Capturing() as output:
            calc_cmd.calculation_list()

        out_str = ''.join(output)
        self.assertTrue(calc_states.TOSUBMIT in out_str, 'TOSUBMIT state not found in: {}'.format(out_str))
        self.assertTrue(calc_states.COMPUTED in out_str, 'COMPUTED state not found in: {}'.format(out_str))
        self.assertFalse(calc_states.FINISHED in out_str, 'FINISHED state not found in: {}'.format(out_str))

        with Capturing() as output:
            calc_cmd.calculation_list(*['-a'])

        out_str = ''.join(output)
        self.assertTrue(calc_states.FINISHED in out_str, 'FINISHED state not found in: {}'.format(out_str))


class TestVerdiCodeCommands(AiidaTestCase):

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create the computers and setup a codes
        """
        super(TestVerdiCodeCommands, cls).setUpClass()

        # Setup computer #1
        from aiida.cmdline.commands.computer import Computer
        cmd_comp = Computer()
        with mock.patch('__builtin__.raw_input', side_effect=computer_setup_input_1):
            with Capturing():
                cmd_comp.computer_setup()

        # Setup a code for computer #1
        from aiida.cmdline.commands.code import Code
        code_cmd = Code()
        with mock.patch('__builtin__.raw_input', side_effect=code_setup_input_1):
            with Capturing():
                cls.code1 = code_cmd.code_setup()

        # Setup computer #2
        with mock.patch('__builtin__.raw_input', side_effect=computer_setup_input_2):
            with Capturing():
                cmd_comp.computer_setup()

        # Setup a code for computer #2
        with mock.patch('__builtin__.raw_input', side_effect=code_setup_input_2):
            with Capturing():
                cls.code2 = code_cmd.code_setup()

    def test_code_list(self):
        """
        Do some code listing test to ensure the correct behaviour of
        verdi code list
        """
        from aiida.cmdline.commands.code import Code
        code_cmd = Code()

        # Run a simple verdi code list, capture the output and check the result
        with Capturing() as output:
            code_cmd.code_list()
        out_str_1 = ''.join(output)
        self.assertTrue(computer_name_1 in out_str_1, "The computer 1 name should be included into this list")
        self.assertTrue(code_name_1 in out_str_1, "The code 1 name should be included into this list")
        self.assertTrue(computer_name_2 in out_str_1, "The computer 2 name should be included into this list")
        self.assertTrue(code_name_2 in out_str_1, "The code 2 name should be included into this list")

        # Run a verdi code list -a, capture the output and check if the result
        # is the same as the previous one
        with Capturing() as output:
            code_cmd.code_list(*['-a'])
        out_str_2 = ''.join(output)
        self.assertEqual(out_str_1, out_str_2, "verdi code list & verdi code list -a should provide "
                         "the same output in this experiment.")

        # Run a verdi code list -c, capture the output and check the result
        with Capturing() as output:
            code_cmd.code_list(*['-c', computer_name_1])
        out_str = ''.join(output)
        self.assertTrue(computer_name_1 in out_str, "The computer 1 name should be included into this list")
        self.assertFalse(computer_name_2 in out_str, "The computer 2 name should not be included into this list")

        # Hide code 2
        self.code2._hide()

        # Run verdi code list again, checking that code 2 is now not shown
        with Capturing() as output:
            code_cmd.code_list()
        out_str_3 = ''.join(output)
        self.assertTrue(computer_name_1 in out_str_3, "The computer 1 name should be included into this list")
        self.assertTrue(code_name_1 in out_str_3, "The code 1 name should be included into this list")
        self.assertFalse(computer_name_2 in out_str_3, "The computer 2 name should not be included into this list")
        self.assertFalse(code_name_2 in out_str_3, "The code 2 name should not be included into this list")


# pylint: disable=abstract-method
class Wf(WorkChain):
    """
    Utility workchain used for testing
    """
    TEST_STRING = 'Test report.'

    @classmethod
    def define(cls, spec):
        super(Wf, cls).define(spec)
        spec.outline(cls.create_logs)

    def create_logs(self):
        self.report(self.TEST_STRING)


class TestVerdiWorkCommands(AiidaTestCase):

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create a simple workchain and run it.
        """
        super(TestVerdiWorkCommands, cls).setUpClass()
        from aiida.work.launch import run_get_pid

        cls.test_string = Wf.TEST_STRING

        _, cls.workchain_pid = run_get_pid(Wf)

    def test_report(self):
        """
        Test that 'verdi work report' contains the report message.
        """
        from aiida.cmdline.commands.work import report

        result = CliRunner().invoke(report, [str(self.workchain_pid)], catch_exceptions=False)
        self.assertTrue(self.test_string in result.output)

    def test_report_debug(self):
        """
        Test that 'verdi work report' contains the report message when called with levelname DEBUG.
        """
        from aiida.cmdline.commands.work import report

        result = CliRunner().invoke(report, [str(self.workchain_pid), '--levelname', 'DEBUG'], catch_exceptions=False)
        self.assertTrue(self.test_string in result.output)

    def test_report_error(self):
        """
        Test that 'verdi work report' does not contain the report message when called with levelname ERROR.
        """
        from aiida.cmdline.commands.work import report

        result = CliRunner().invoke(report, [str(self.workchain_pid), '--levelname', 'ERROR'], catch_exceptions=False)
        self.assertTrue(self.test_string not in result.output)


# pylint: disable=no-self-use
class TestVerdiUserCommands(AiidaTestCase):

    def setUp(self):
        """
        Create a user
        """
        super(TestVerdiUserCommands, self).setUp()

        # Setup user #1
        from aiida.cmdline.commands.user import do_configure

        with mock.patch('__builtin__.raw_input', side_effect=computer_setup_input_1):
            with Capturing():
                do_configure(
                    self.backend,
                    user_1['email'],
                    user_1['first_name'],
                    user_1['last_name'],
                    user_1['institution'],
                    no_password=True,
                    non_interactive=True,
                    force_reconfigure=True)

    def test_user_list(self):
        """
        verdi user list
        """
        from aiida.cmdline.commands.user import list as list_user

        result = CliRunner().invoke(list_user, [], catch_exceptions=False)
        self.assertTrue(user_1['email'] in result.output)

    def test_user_configure(self):
        """
        Try configuring a new user
        verdi user configure
        """
        from aiida.cmdline.commands.user import configure

        cli_options = [
            user_2['email'], '--first-name', user_2['first_name'], '--last-name', user_2['last_name'], '--institution',
            user_2['institution'], '--no-password', '--force-reconfigure'
        ]

        # configure user
        result = CliRunner().invoke(configure, cli_options, catch_exceptions=False)
        self.assertTrue(user_2['email'] in result.output)
        self.assertTrue("is already present" not in result.output)

        # reconfigure user
        result = CliRunner().invoke(configure, cli_options, catch_exceptions=False)
        self.assertTrue(user_2['email'] in result.output)
        self.assertTrue("is already present" in result.output)


class TestVerdiDataCommands(AiidaTestCase):
    cmd_to_nodeid_map = dict()
    cmd_to_nodeid_map_for_groups = dict()
    cmd_to_nodeid_map_for_nuser = dict()

    group_name = "trj_group"
    group_id = None

    @classmethod
    def create_trajectory_data(cls, cmd_to_nodeid_map, cmd_to_nodeid_map_for_groups, cmd_to_nodeid_map_for_nuser, group,
                               new_user):

        from aiida.orm.data.array.trajectory import TrajectoryData
        from aiida.cmdline.commands.data import _Trajectory
        import numpy

        # Create the Trajectory data nodes
        tjn1 = TrajectoryData()

        # I create sample data
        stepids = numpy.array([60, 70])
        times = stepids * 0.01
        cells = numpy.array([[[2., 0., 0.], [0., 2., 0.], [0., 0., 2.]], [[3., 0., 0.], [0., 3., 0.], [0., 0., 3.]]])
        symbols = numpy.array(['H', 'O', 'C'])
        positions = numpy.array([[[0., 0., 0.], [0.5, 0.5, 0.5], [1.5, 1.5, 1.5]], [[0., 0., 0.], [0.5, 0.5, 0.5],
                                                                                    [1.5, 1.5, 1.5]]])
        velocities = numpy.array([[[0., 0., 0.], [0., 0., 0.], [0., 0., 0.]], [[0.5, 0.5, 0.5], [0.5, 0.5, 0.5],
                                                                               [-0.5, -0.5, -0.5]]])

        # I set the node
        tjn1.set_trajectory(
            stepids=stepids, cells=cells, symbols=symbols, positions=positions, times=times, velocities=velocities)
        tjn1.store()

        tjn2 = TrajectoryData()
        tjn2.set_trajectory(
            stepids=stepids, cells=cells, symbols=symbols, positions=positions, times=times, velocities=velocities)
        tjn2.store()

        # Keep track of the created objects
        cmd_to_nodeid_map[_Trajectory] = [tjn1.id, tjn2.id]

        # Add the second Trajectory data to the group
        group.add_nodes([tjn2])
        # Keep track of the id of the node that you added to the group
        cmd_to_nodeid_map_for_groups[_Trajectory] = tjn2.id

        # Create a trajectory data that belongs to another user
        tjn3 = TrajectoryData()
        tjn3.set_trajectory(
            stepids=stepids, cells=cells, symbols=symbols, positions=positions, times=times, velocities=velocities)
        tjn3.set_user(new_user)
        tjn3.store()

        # Put it is to the right map
        cmd_to_nodeid_map_for_nuser[_Trajectory] = [tjn3.id]

    @classmethod
    def create_cif_data(cls, cmd_to_nodeid_map, cmd_to_nodeid_map_for_groups, cmd_to_nodeid_map_for_nuser, group,
                        new_user):

        from aiida.orm.data.cif import CifData
        from aiida.cmdline.commands.data import _Cif
        import tempfile

        # Create the CIF data nodes
        with tempfile.NamedTemporaryFile() as f:
            f.write('''
                 data_9012064
                 _space_group_IT_number           166
                 _symmetry_space_group_name_H-M   'R -3 m :H'
                 _cell_angle_alpha                90
                 _cell_angle_beta                 90
                 _cell_angle_gamma                120
                 _cell_length_a                   4.395
                 _cell_length_b                   4.395
                 _cell_length_c                   30.440
                 _cod_database_code               9012064
                 loop_
                 _atom_site_label
                 _atom_site_fract_x
                 _atom_site_fract_y
                 _atom_site_fract_z
                 _atom_site_U_iso_or_equiv
                 Bi 0.00000 0.00000 0.40046 0.02330
                 Te1 0.00000 0.00000 0.00000 0.01748
                 Te2 0.00000 0.00000 0.79030 0.01912
             ''')
            f.flush()
            c1 = CifData(file=f.name)
            c1.store()
            c2 = CifData(file=f.name)
            c2.store()

            # Keep track of the created objects
            cmd_to_nodeid_map[_Cif] = [c1.id, c2.id]

            # Add the second CIF data to the group
            group.add_nodes([c2])
            # Keep track of the id of the node that you added to the group
            cmd_to_nodeid_map_for_groups[_Cif] = c2.id

            # Create a Cif node belonging to another user
            c3 = CifData(file=f.name)
            c3.set_user(new_user)
            c3.store()

            # Put it is to the right map
            cmd_to_nodeid_map_for_nuser[_Cif] = [c3.id]

    @classmethod
    def sub_create_bands_data(cls, user=None):
        from aiida.orm.data.array.kpoints import KpointsData
        from aiida.orm import JobCalculation
        from aiida.orm.data.structure import StructureData
        from aiida.common.links import LinkType
        from aiida.orm.data.array.bands import BandsData
        import numpy

        s = StructureData(cell=((2., 0., 0.), (0., 2., 0.), (0., 0., 2.)))
        s.append_atom(position=(0., 0., 0.), symbols=['Ba', 'Ti'], weights=(1., 0.), name='mytype')
        if user is not None:
            s.set_user(user)
        s.store()

        c = JobCalculation(computer=cls.computer, resources={'num_machines': 1, 'num_mpiprocs_per_machine': 1})
        if user is not None:
            c.set_user(user)
        c.store()
        c.add_link_from(s, "S1", LinkType.INPUT)
        c._set_state(calc_states.RETRIEVING)

        # define a cell
        alat = 4.
        cell = numpy.array([
            [alat, 0., 0.],
            [0., alat, 0.],
            [0., 0., alat],
        ])

        k = KpointsData()
        k.set_cell(cell)
        k.set_kpoints_path()
        if user is not None:
            k.set_user(user)
        k.store()

        b = BandsData()
        b.set_kpointsdata(k)
        input_bands = numpy.array([numpy.ones(4) * i for i in range(k.get_kpoints().shape[0])])
        b.set_bands(input_bands, units='eV')
        if user is not None:
            b.set_user(user)
        b.store()

        b.add_link_from(c, link_type=LinkType.CREATE)

        return b

    @classmethod
    def create_bands_data(cls, cmd_to_nodeid_map, cmd_to_nodeid_map_for_groups, cmd_to_nodeid_map_for_nuser, group,
                          new_user):
        from aiida.cmdline.commands.data import _Bands

        b1 = cls.sub_create_bands_data()
        b2 = cls.sub_create_bands_data()

        # Keep track of the created objects
        cmd_to_nodeid_map[_Bands] = [b1.id, b2.id]

        # Add the second Kpoint & Bands data to the group
        group.add_nodes([b2])
        # Keep track of the id of the node that you added to the group
        cmd_to_nodeid_map_for_groups[_Bands] = b2.id

        b3 = cls.sub_create_bands_data(user=new_user)
        # Put it is to the right map (of the different user)
        cmd_to_nodeid_map_for_nuser[_Bands] = [b3.id]

    @classmethod
    def create_structure_data(cls, cmd_to_nodeid_map, cmd_to_nodeid_map_for_groups, cmd_to_nodeid_map_for_nuser, group,
                              new_user):
        from aiida.orm.data.structure import StructureData
        from aiida.cmdline.commands.data import _Structure

        s1 = StructureData(cell=((2., 0., 0.), (0., 2., 0.), (0., 0., 2.)))
        s1.append_atom(position=(0., 0., 0.), symbols=['Ba', 'Ti'], weights=(1., 0.), name='mytype')
        s1.store()

        s2 = StructureData(cell=((2., 0., 0.), (0., 2., 0.), (0., 0., 2.)))
        s2.append_atom(position=(0., 0., 0.), symbols=['Ba', 'Ti'], weights=(1., 0.), name='mytype')
        s2.store()

        # Keep track of the created objects
        cmd_to_nodeid_map[_Structure] = [s1.id, s2.id]

        # Add the second Kpoint & Bands data to the group
        group.add_nodes([s2])
        # Keep track of the id of the node that you added to the group
        cmd_to_nodeid_map_for_groups[_Structure] = s2.id

        # Create a StructureData node belonging to another user
        s3 = StructureData(cell=((2., 0., 0.), (0., 2., 0.), (0., 0., 2.)))
        s3.append_atom(position=(0., 0., 0.), symbols=['Ba', 'Ti'], weights=(1., 0.), name='mytype')
        s3.set_user(new_user)
        s3.store()

        # Put it is to the right map
        cmd_to_nodeid_map_for_nuser[_Structure] = [s3.id]

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create some data needed for the tests
        """
        super(TestVerdiDataCommands, cls).setUpClass()

        from aiida.orm.group import Group

        # Create a secondary user
        new_email = "newuser@new.n"
        new_user = cls.backend.users.create(email=new_email)
        new_user.store()

        # Create a group to add specific data inside
        g1 = Group(name=cls.group_name)
        g1.store()
        cls.group_id = g1.id

        cls.create_bands_data(cls.cmd_to_nodeid_map, cls.cmd_to_nodeid_map_for_groups, cls.cmd_to_nodeid_map_for_nuser,
                              g1, new_user)

        cls.create_structure_data(cls.cmd_to_nodeid_map, cls.cmd_to_nodeid_map_for_groups,
                                  cls.cmd_to_nodeid_map_for_nuser, g1, new_user)

        cls.create_cif_data(cls.cmd_to_nodeid_map, cls.cmd_to_nodeid_map_for_groups, cls.cmd_to_nodeid_map_for_nuser,
                            g1, new_user)

        cls.create_trajectory_data(cls.cmd_to_nodeid_map, cls.cmd_to_nodeid_map_for_groups,
                                   cls.cmd_to_nodeid_map_for_nuser, g1, new_user)

    def test_trajectory_simple_listing(self):
        from aiida.cmdline.commands.data import _Bands
        from aiida.cmdline.commands.data import _Structure
        from aiida.cmdline.commands.data import _Cif
        from aiida.cmdline.commands.data import _Trajectory

        sub_cmds = [_Bands, _Structure, _Cif, _Trajectory]
        for sub_cmd in sub_cmds:
            with Capturing() as output:
                sub_cmd().list()

            out_str = ' '.join(output)

            for nid in self.cmd_to_nodeid_map[sub_cmd]:
                if str(nid) not in out_str:
                    self.fail("The data objects ({}) with ids {} and {} "
                              "were not found. "
                              .format(sub_cmd, str(self.cmd_to_nodeid_map[sub_cmd][0]),
                                      str(self.cmd_to_nodeid_map[sub_cmd][1])) + "The output was {}".format(out_str))

    def test_trajectory_all_user_listing(self):
        from aiida.cmdline.commands.data import _Bands
        from aiida.cmdline.commands.data import _Structure
        from aiida.cmdline.commands.data import _Cif
        from aiida.cmdline.commands.data import _Trajectory

        sub_cmds = [_Bands, _Structure, _Cif, _Trajectory]
        for sub_cmd in sub_cmds:
            args_to_test = [['-A'], ['--all-users']]
            for arg in args_to_test:
                curr_scmd = sub_cmd()
                with Capturing() as output:
                    curr_scmd.list(*arg)

                out_str = ' '.join(output)

                for nid in self.cmd_to_nodeid_map[sub_cmd] + self.cmd_to_nodeid_map_for_nuser[sub_cmd]:
                    if str(nid) not in out_str:
                        self.fail("The data objects ({}) with ids {} and {} "
                                  "were not found. ".format(sub_cmd, str(self.cmd_to_nodeid_map[sub_cmd][0]),
                                                            str(self.cmd_to_nodeid_map[sub_cmd][1])) +
                                  "The output was {}".format(out_str))

    def test_trajectory_past_days_listing(self):
        from aiida.cmdline.commands.data import _Bands
        from aiida.cmdline.commands.data import _Structure
        from aiida.cmdline.commands.data import _Cif
        from aiida.cmdline.commands.data import _Trajectory

        sub_cmds = [_Bands, _Structure, _Cif, _Trajectory]
        for sub_cmd in sub_cmds:
            args_to_test = [['-p', '0'], ['--past-days', '0']]
            for arg in args_to_test:
                curr_scmd = sub_cmd()
                with Capturing() as output:
                    curr_scmd.list(*arg)
                out_str = ' '.join(output)

                # This should be an empty output
                for nid in self.cmd_to_nodeid_map[sub_cmd]:
                    if str(nid) in out_str:
                        self.fail("No data objects should be retrieved and "
                                  "some were retrieved. The (concatenation of "
                                  "the) output was: {}".format(out_str))

            args_to_test = [['-p', '1'], ['--past-days', '1']]
            for arg in args_to_test:
                curr_scmd = sub_cmd()
                with Capturing() as output:
                    curr_scmd.list(*arg)
                out_str = ' '.join(output)

                for nid in self.cmd_to_nodeid_map[sub_cmd]:
                    if str(nid) not in out_str:
                        self.fail("The data objects ({}) with ids {} and {} "
                                  "were not found. ".format(sub_cmd, str(self.cmd_to_nodeid_map[sub_cmd][0]),
                                                            str(self.cmd_to_nodeid_map[sub_cmd][1])) +
                                  "The output was {}".format(out_str))

    def test_trajectory_group_listing(self):
        from aiida.cmdline.commands.data import _Bands
        from aiida.cmdline.commands.data import _Structure
        from aiida.cmdline.commands.data import _Cif
        from aiida.cmdline.commands.data import _Trajectory

        args_to_test = [['-g', self.group_name], ['--group-name', self.group_name], ['-G', str(self.group_id)],
                        ['--group-pk', str(self.group_id)]]

        sub_cmds = [_Bands, _Structure, _Cif, _Trajectory]
        for sub_cmd in sub_cmds:
            for arg in args_to_test:
                curr_scmd = sub_cmd()
                with Capturing() as output:
                    curr_scmd.list(*arg)
                out_str = ' '.join(output)

                if str(self.cmd_to_nodeid_map_for_groups[sub_cmd]) not in out_str:
                    self.fail(
                        "The data object ({}) with id {} "
                        "was not found. ".format(
                            sub_cmd,
                            str(self.cmd_to_nodeid_map_for_groups[sub_cmd]) + "The output was {}".format(out_str)))


class TestVerdiDataRemoteCommands(AiidaTestCase):
    """
    Test the commands under 'verdi data remote'

    Implicitly also tests creating and configuring a computer with a local transport
    """

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create a configured computer to
        """
        from aiida.orm import Computer
        from aiida.cmdline.commands.computer import Computer as ComputerCmd
        from aiida.orm.backend import construct_backend

        super(TestVerdiDataRemoteCommands, cls).setUpClass()

        backend = construct_backend()

        cls.computer_name = 'test_remote_ls'
        cls.new_comp = Computer(
            name=cls.computer_name,
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida')
        cls.new_comp.store()

        # I need to configure the computer here; being 'local',
        # there should not be any options asked here
        with Capturing():
            ComputerCmd().run('configure', cls.computer_name)

        assert cls.new_comp.is_user_configured(
            backend.users.get_automatic_user()), "There was a problem configuring the test computer"

    def test_remote_ls(self):
        """
        Test if the 'verdi remote ls' command works
        """
        from aiida.cmdline.commands.data import _Remote
        import os
        from aiida.orm.data.remote import RemoteData
        from aiida.common.folders import SandboxFolder

        with SandboxFolder() as folder:
            files = {'test1.txt': 'the_content_1', 'test2.txt': 'the_content_2'}
            for fname, content in files.items():
                with open(os.path.join(folder.abspath, fname), 'w') as f:
                    f.write(content)

            r = RemoteData(computer=self.new_comp, remote_path=folder.abspath)
            r.store()

            with Capturing() as output:
                _Remote().run('ls', str(r.pk))

            # output is a Capturing objects, looping on it returns the lines
            found_files = set(output)
            self.assertEquals(set(files.keys()), found_files)

            # Testing also ls -l, that calls a different implementation
            with Capturing() as output:
                _Remote().run('ls', '-l', str(r.pk))

            # The filename is the last part of each line
            found_files = set(_.split()[-1] for _ in output)
            self.assertEquals(set(files.keys()), found_files)

    def test_remote_cat(self):
        """
        Test if the 'verdi remote ls' command works
        """
        from aiida.cmdline.commands.data import _Remote
        import os
        from aiida.orm.data.remote import RemoteData
        from aiida.common.folders import SandboxFolder

        with SandboxFolder() as folder:

            files = {'test1.txt': 'the_content_1\nsecond line', 'test2.txt': 'the_content_2'}
            for fname, content in files.items():
                with open(os.path.join(folder.abspath, fname), 'w') as f:
                    f.write(content)

            r = RemoteData(computer=self.new_comp, remote_path=folder.abspath)
            r.store()

            for fname, content in files.items():
                with Capturing() as output:
                    _Remote().run('cat', str(r.pk), fname)

                self.assertEquals("\n".join(output), content, "The file content for file {} differs: {} vs. {}".format(
                    fname, "\n".join(output), content))


class TestVerdiComputerCommands(AiidaTestCase):
    """
    Test the commands under 'verdi computer'

    Implicitly also tests creating and configuring a computer with a local transport
    """

    @classmethod
    def setUpClass(cls, *args, **kwargs):
        """
        Create a configured computer to
        """
        from aiida.orm import Computer
        from aiida.cmdline.commands.computer import Computer as ComputerCmd
        from aiida.orm.backend import construct_backend

        super(TestVerdiComputerCommands, cls).setUpClass()

        backend = construct_backend()
        cls.computer_name = 'test_computer'
        cls.new_comp = Computer(
            name=cls.computer_name,
            hostname='localhost',
            transport_type='local',
            scheduler_type='direct',
            workdir='/tmp/aiida')
        cls.new_comp.store()

        # I need to configure the computer here; being 'local',
        # there should not be any options asked here
        with Capturing():
            ComputerCmd().run('configure', cls.computer_name)

        assert cls.new_comp.is_user_configured(
            backend.users.get_automatic_user()), "There was a problem configuring the test computer"

    def test_computer_test(self):
        """
        Test if the 'verdi computer test' command works

        It should work as it is a local connection
        """
        from aiida.cmdline.commands.computer import Computer as ComputerCmd

        # Check that indeed, if there is a problem, we detect it as such
        with self.assertRaises(SystemExit):
            with Capturing(capture_stderr=True):
                ComputerCmd().run('test', "not_existent_computer_name")

        # Test the computer
        with Capturing():
            ComputerCmd().run('test', self.computer_name)

    def test_computer_enable_disable(self):
        """
        Check if the computer enable/disable works
        """
        from aiida.cmdline.commands.computer import Computer as ComputerCmd
        from aiida.orm.backend import construct_backend

        backend = construct_backend()

        user = backend.users.get_automatic_user()
        self.assertTrue(self.new_comp.is_enabled())
        self.assertTrue(self.new_comp.is_user_enabled(user))

        with Capturing():
            ComputerCmd().run('disable', '-u', user.email, self.computer_name)
        self.assertFalse(self.new_comp.is_user_enabled(user))

        with Capturing():
            ComputerCmd().run('enable', '-u', user.email, self.computer_name)
        self.assertTrue(self.new_comp.is_user_enabled(user))

        try:
            with Capturing():
                ComputerCmd().run('disable', self.computer_name)

            self.assertFalse(self.new_comp.is_enabled())

            with Capturing():
                ComputerCmd().run('enable', self.computer_name)
            self.assertTrue(self.new_comp.is_enabled())
        finally:
            # Make sure we re-enable even if some assertions are false
            # as this can potentially block the tests (even if I don't know why)
            with Capturing():
                ComputerCmd().run('enable', self.computer_name)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=cell-var-from-loop
"""Convenience classes to help building the input dictionaries for Processes."""
from collections import Mapping
from aiida.work.ports import PortNamespace

__all__ = ['ProcessBuilder', 'JobProcessBuilder', 'ProcessBuilderNamespace']


class ProcessBuilderNamespace(Mapping):
    """
    Input namespace for the ProcessBuilder. Dynamically generates the getters and setters
    for the input ports of a given PortNamespace
    """

    def __init__(self, port_namespace):
        """
        Dynamically construct the get and set properties for the ports of the given port namespace

        For each port in the given port namespace a get and set property will be constructed dynamically
        and added to the ProcessBuilderNamespace. The docstring for these properties will be defined
        by calling str() on the Port, which should return the description of the Port.

        :param port_namespace: the inputs PortNamespace for which to construct the builder
        """
        # pylint: disable=super-init-not-called
        self._port_namespace = port_namespace
        self._valid_fields = []
        self._data = {}

        for name, port in port_namespace.items():

            self._valid_fields.append(name)

            if isinstance(port, PortNamespace):
                self._data[name] = ProcessBuilderNamespace(port)

                def fgetter(self, name=name):
                    return self._data.get(name)
            elif port.has_default():

                def fgetter(self, name=name, default=port.default):
                    return self._data.get(name, default)
            else:

                def fgetter(self, name=name):
                    return self._data.get(name, None)

            def fsetter(self, value):
                self._data[name] = value

            fgetter.__doc__ = str(port)
            getter = property(fgetter)
            getter.setter(fsetter)
            setattr(self.__class__, name, getter)

    def __setattr__(self, attr, value):
        """
        Any attributes without a leading underscore being set correspond to inputs and should hence
        be validated with respect to the corresponding input port from the process spec
        """
        if attr.startswith('_'):
            object.__setattr__(self, attr, value)
        else:
            try:
                port = self._port_namespace[attr]
            except KeyError:
                raise AttributeError('Unknown builder parameter: {}'.format(attr))

            value = port.serialize(value)
            is_valid, message = port.validate(value)
            if not is_valid:
                raise ValueError('invalid attribute value: {}'.format(message))

            self._data[attr] = value

    def __repr__(self):
        return self._data.__repr__()

    def __dir__(self):
        return sorted(set(self._valid_fields + [key for key, _ in self.__dict__ if key.startswith('_')]))

    def __iter__(self):
        for key in self._data:
            yield key

    def __len__(self):
        return len(self._data)

    def __getitem__(self, item):
        return self._data[item]


class ProcessBuilder(ProcessBuilderNamespace):
    """
    A process builder that helps creating a new calculation
    """

    def __init__(self, process_class):
        self._process_class = process_class
        self._process_spec = self._process_class.spec()
        super(ProcessBuilder, self).__init__(port_namespace=self._process_spec.inputs)

    @property
    def process_class(self):
        return self._process_class


class JobProcessBuilder(ProcessBuilder):
    """
    A process builder specific to JobCalculation classes, that provides
    also the submit_test functionality
    """

    def __dir__(self):
        return super(JobProcessBuilder, self).__dir__() + ['submit_test']

    def submit_test(self, folder=None, subfolder_name=None):
        """
        Run a test submission by creating the files that would be generated for the real calculation in a local folder,
        without actually storing the calculation nor the input nodes. This functionality therefore also does not
        require any of the inputs nodes to be stored yet.

        :param folder: a Folder object, within which to create the calculation files. By default a folder
            will be created in the current working directory
        :param subfolder_name: the name of the subfolder to use within the directory of the ``folder`` object. By
            default a unique string will be generated based on the current datetime with the format ``yymmdd-``
            followed by an auto incrementing index
        """
        inputs = {'store_provenance': False}
        inputs.update(**self)
        process = self._process_class(inputs=inputs)

        return process.calc.submit_test(folder, subfolder_name)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida_core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
from aiida import load_dbenv, is_dbenv_loaded

if not is_dbenv_loaded():
    load_dbenv()


from aiida import work
from aiida.orm.data.int import Int
import aiida.work.globals as globals

persistence = globals.get_persistence()


class ProdSum(work.WorkChain):
    @classmethod
    def define(cls, spec):
        super(ProdSum, cls).define(spec)
        spec.input("a")
        spec.input("b")
        spec.input("c")
        spec.outline(cls.add, cls.prod)

    def add(self):
        self.ctx.sum = self.inputs.a + self.inputs.b

    def prod(self):
        result = self.ctx.sum * self.inputs.c
        self.out("result", result)


class ProdSumBroken(ProdSum):
    def prod(self):
        raise RuntimeError("Break")


p = ProdSumBroken(inputs={'a': Int(2), 'b': Int(3), 'c': Int(4)})
pid = p.pid
persistence.persist_process(p)
try:
    p.play()
except RuntimeError:
    pass

cp = persistence.load_checkpoint(pid)
p = ProdSum.recreate_from(cp)

p.play()

print(p.outputs)
